from typing import Literal


class Evo2Config:
    """Configuration class for Evo 2 models.

    Parameters
    ----------
    model_name : str, optional
        The name of the model to use. The default is "evo2-7b". Options are "evo2-7b", "evo2-7b-base", "evo2-1b-base".
    batch_size : int, optional
        The batch size to use. The default is 1.
    """

    def __init__(
        self,
        model_name: Literal["evo2-7b", "evo2-7b-base", "evo2-1b-base"] = "evo2-7b",
        batch_size: int = 1,
    ):
        # model specific parameters
        self.model_map = {
            "evo2-7b": {
                "model_name": "evo2_7b",
                "model_hf_name": "arcinstitute/evo2_7b",
                "default_embedding_layer": "blocks.31.mlp.l3",
                "vocab_size": 512,
                "hidden_size": 4096,
                # Number of long convolution filters in each hyena block. Can be smaller than `hidden_size`
                "num_filters": 4096,
                "hcl_layer_idxs": [2, 6, 9, 13, 16, 20, 23, 27, 30],
                "hcm_layer_idxs": [1, 5, 8, 12, 15, 19, 22, 26, 29],
                "hcs_layer_idxs": [0, 4, 7, 11, 14, 18, 21, 25, 28],
                "attn_layer_idxs": [3, 10, 17, 24, 31],
                "hcm_filter_length": 128,
                "hcl_filter_groups": 4096,
                "hcm_filter_groups": 256,
                "hcs_filter_groups": 256,
                "hcs_filter_length": 7,
                "num_layers": 32,
                # Length of the short, depthwise FIR applied to input projections
                "short_filter_length": 3,
                "num_attention_heads": 32,
                "short_filter_bias": False,  # add bias to FIR
                "mlp_init_method": "torch.nn.init.zeros_",
                "mlp_output_init_method": "torch.nn.init.zeros_",
                "eps": 0.000001,
                "state_size": 16,
                "rotary_emb_base": 100000000000,
                "rotary_emb_scaling_factor": 128,
                "use_interpolated_rotary_pos_emb": True,
                "make_vocab_size_divisible_by": 8,
                "inner_size_multiple_of": 16,  # force GLU inner_size to be a multiple of
                "inner_mlp_size": 11264,
                "log_intermediate_values": False,
                # Number of groups in GQA
                "proj_groups": 1,
                # Number of groups in grouped
                "hyena_filter_groups": 1,
                # Split strategy for channels
                "column_split_hyena": False,
                "column_split": True,
                "interleave": True,
                # Layer > 0 nn.identity activation
                "evo2_style_activations": True,
                # Legacy options for MP / PP inference
                "model_parallel_size": 1,
                "pipe_parallel_size": 1,
                "tie_embeddings": True,
                "mha_out_proj_bias": True,
                "hyena_out_proj_bias": True,
                "hyena_flip_x1x2": False,
                "qkv_proj_bias": False,
                "use_fp8_input_projections": True,
                "max_seqlen": 1048576,
                "final_norm": True,
                "use_flash_attn": True,
                "use_flash_rmsnorm": False,
                "use_flash_depthwise": False,
                "use_flashfft": False,
                "use_laughing_hyena": False,
                "inference_mode": True,
                "prefill_style": "fft",
                "mlp_activation": "gelu",
                "print_activations": False,
            },
            "evo2-7b-base": {
                "model_name": "evo2_7b_base",
                "model_hf_name": "arcinstitute/evo2_7b_base",
                "default_embedding_layer": "blocks.31.mlp.l3",
                "vocab_size": 512,
                "hidden_size": 4096,
                "num_filters": 4096,
                "hcl_layer_idxs": [2, 6, 9, 13, 16, 20, 23, 27, 30],
                "hcm_layer_idxs": [1, 5, 8, 12, 15, 19, 22, 26, 29],
                "hcs_layer_idxs": [0, 4, 7, 11, 14, 18, 21, 25, 28],
                "attn_layer_idxs": [3, 10, 17, 24, 31],
                "hcm_filter_length": 128,
                "hcl_filter_groups": 4096,
                "hcm_filter_groups": 256,
                "hcs_filter_groups": 256,
                "hcs_filter_length": 7,
                "num_layers": 32,
                "short_filter_length": 3,
                "num_attention_heads": 32,
                "short_filter_bias": False,
                "mlp_init_method": "torch.nn.init.zeros_",
                "mlp_output_init_method": "torch.nn.init.zeros_",
                "eps": 0.000001,
                "state_size": 16,
                "rotary_emb_base": 10000,
                "make_vocab_size_divisible_by": 8,
                "inner_size_multiple_of": 16,
                "inner_mlp_size": 11008,
                "log_intermediate_values": False,
                "proj_groups": 1,
                "hyena_filter_groups": 1,
                "column_split_hyena": False,
                "column_split": True,
                "interleave": True,
                "evo2_style_activations": True,
                "model_parallel_size": 1,
                "pipe_parallel_size": 1,
                "tie_embeddings": True,
                "mha_out_proj_bias": True,
                "hyena_out_proj_bias": True,
                "hyena_flip_x1x2": False,
                "qkv_proj_bias": False,
                "use_fp8_input_projections": False,
                "max_seqlen": 32768,
                "max_batch_size": 1,
                "final_norm": True,
                "use_flash_attn": True,
                "use_flash_rmsnorm": False,
                "use_flash_depthwise": False,
                "use_flashfft": False,
                "use_laughing_hyena": False,
                "inference_mode": True,
                "tokenizer_type": "CharLevelTokenizer",
                "prefill_style": "fft",
                "mlp_activation": "gelu",
                "print_activations": False,
                "use_interpolated_rotary_pos_emb": True,
            },
            "evo2-1b-base": {
                "model_name": "evo2_1b_base",
                "model_hf_name": "arcinstitute/evo2_1b_base",
                "default_embedding_layer": "blocks.24.mlp.l3",
                "vocab_size": 512,
                "hidden_size": 1920,
                # Number of independent filters in Hyena-LI
                "num_filters": 1920,
                "attn_layer_idxs": [3, 10, 17, 24],
                "hcl_layer_idxs": [2, 6, 9, 13, 16, 20, 23],
                "hcm_layer_idxs": [1, 5, 8, 12, 15, 19, 22],
                "hcs_layer_idxs": [0, 4, 7, 11, 14, 18, 21],
                "hcm_filter_length": 128,
                "hcl_filter_groups": 1920,
                "hcm_filter_groups": 128,
                "hcs_filter_groups": 128,
                "hcs_filter_length": 7,
                "num_layers": 25,
                # Length of the short, depthwise FIR applied to input projections
                "short_filter_length": 3,
                "num_attention_heads": 15,
                "short_filter_bias": False,  # add bias to FIR
                "mlp_init_method": "torch.nn.init.zeros_",
                "mlp_output_init_method": "torch.nn.init.zeros_",
                "eps": 0.000001,
                "state_size": 16,
                "rotary_emb_base": 10000,
                "make_vocab_size_divisible_by": 8,
                "inner_size_multiple_of": 16,  # force GLU inner_size to be a multiple of
                "inner_mlp_size": 5120,
                "log_intermediate_values": False,
                # Number of groups in GQA
                "proj_groups": 1,
                # Number of groups in grouped
                "hyena_filter_groups": 1,
                # Split strategy for channels
                "column_split_hyena": False,
                "column_split": True,
                "interleave": True,
                # Layer > 0 nn.identity activation
                "evo2_style_activations": True,
                # Legacy options for MP / PP inference
                "model_parallel_size": 1,
                "pipe_parallel_size": 1,
                "tie_embeddings": True,
                "mha_out_proj_bias": True,
                "hyena_out_proj_bias": True,
                "hyena_flip_x1x2": False,
                "qkv_proj_bias": False,
                "use_fp8_input_projections": True,
                "max_seqlen": 8192,
                "final_norm": True,
                "use_flash_attn": True,
                "use_flash_rmsnorm": False,
                "use_flash_depthwise": False,
                "use_flashfft": False,
                "use_laughing_hyena": False,
                "inference_mode": True,
                "prefill_style": "fft",
                "mlp_activation": "gelu",
                "print_activations": False,
            },
            "evo2-40b-base": {
                "model_name": "evo2_40b_base",
                "model_hf_name": "arcinstitute/evo2_40b_base",
                "default_embedding_layer": "blocks.49.mlp.l3",
                "vocab_size": 512,
                "hidden_size": 8192,
                "num_filters": 8192,
                "hcl_layer_idxs": [2, 6, 9, 13, 16, 20, 23, 27, 30, 34, 38, 41, 45, 48],
                "hcm_layer_idxs": [1, 5, 8, 12, 15, 19, 22, 26, 29, 33, 37, 40, 44, 47],
                "hcs_layer_idxs": [0, 4, 7, 11, 14, 18, 21, 25, 28, 32, 36, 39, 43, 46],
                "attn_layer_idxs": [3, 10, 17, 24, 31, 35, 42, 49],
                "hcm_filter_length": 128,
                "hcl_filter_groups": 8192,
                "hcm_filter_groups": 512,
                "hcs_filter_groups": 512,
                "hcs_filter_length": 7,
                "num_layers": 50,
                "short_filter_length": 3,
                "num_attention_heads": 64,
                "short_filter_bias": False,
                "mlp_init_method": "torch.nn.init.zeros_",
                "mlp_output_init_method": "torch.nn.init.zeros_",
                "eps": 0.000001,
                "state_size": 16,
                "rotary_emb_base": 1000000,
                "make_vocab_size_divisible_by": 8,
                "inner_size_multiple_of": 128,
                "inner_mlp_size": 21888,
                "log_intermediate_values": False,
                "proj_groups": 1,
                "hyena_filter_groups": 1,
                "column_split_hyena": False,
                "column_split": True,
                "interleave": True,
                "evo2_style_activations": True,
                "use_fp8_input_projections": True,
                "model_parallel_size": 1,
                "pipe_parallel_size": 1,
                "tie_embeddings": True,
                "mha_out_proj_bias": True,
                "hyena_out_proj_bias": True,
                "hyena_flip_x1x2": False,
                "qkv_proj_bias": False,
                "max_seqlen": 8192,
                "max_batch_size": 1,
                "final_norm": True,
                "use_flash_attn": True,
                "use_flash_rmsnorm": False,
                "use_flash_depthwise": False,
                "use_flashfft": False,
                "use_laughing_hyena": False,
                "inference_mode": True,
                "prefill_style": "fft",
                "mlp_activation": "gelu",
                "print_activations": False,
            },
            "evo2-40b": {
                "model_name": "evo2_40b",
                "model_hf_name": "arcinstitute/evo2_40b",
                "default_embedding_layer": "blocks.49.mlp.l3",
                "vocab_size": 512,
                "hidden_size": 8192,
                "num_filters": 8192,
                "hcl_layer_idxs": [2, 6, 9, 13, 16, 20, 23, 27, 30, 34, 38, 41, 45, 48],
                "hcm_layer_idxs": [1, 5, 8, 12, 15, 19, 22, 26, 29, 33, 37, 40, 44, 47],
                "hcs_layer_idxs": [0, 4, 7, 11, 14, 18, 21, 25, 28, 32, 36, 39, 43, 46],
                "attn_layer_idxs": [3, 10, 17, 24, 31, 35, 42, 49],
                "hcm_filter_length": 128,
                "hcl_filter_groups": 8192,
                "hcm_filter_groups": 512,
                "hcs_filter_groups": 512,
                "hcs_filter_length": 7,
                "num_layers": 50,
                "short_filter_length": 3,
                "num_attention_heads": 64,
                "short_filter_bias": False,
                "mlp_init_method": "torch.nn.init.zeros_",
                "mlp_output_init_method": "torch.nn.init.zeros_",
                "eps": 0.000001,
                "state_size": 16,
                "rotary_emb_base": 100000000000,
                "rotary_emb_scaling_factor": 128,
                "use_interpolated_rotary_pos_emb": True,
                "make_vocab_size_divisible_by": 8,
                "inner_size_multiple_of": 128,  # force GLU inner_size to be a multiple of
                "inner_mlp_size": 22528,
                "log_intermediate_values": False,
                "proj_groups": 1,
                "hyena_filter_groups": 1,
                "column_split_hyena": False,
                "column_split": True,
                "interleave": True,
                "evo2_style_activations": True,
                "use_fp8_input_projections": True,
                "model_parallel_size": 1,
                "pipe_parallel_size": 1,
                "tie_embeddings": True,
                "mha_out_proj_bias": True,
                "hyena_out_proj_bias": True,
                "hyena_flip_x1x2": False,
                "qkv_proj_bias": False,
                "max_seqlen": 1048576,
                "max_batch_size": 1,
                "final_norm": True,
                "use_flash_attn": True,
                "use_flash_rmsnorm": False,
                "use_flash_depthwise": False,
                "use_flashfft": False,
                "use_laughing_hyena": False,
                "inference_mode": True,
                "prefill_style": "fft",
                "mlp_activation": "gelu",
                "print_activations": False,
            },
        }
        if model_name not in self.model_map:
            raise ValueError(
                f"Model name {model_name} not found in available models: {self.model_map.keys()}"
            )

        self.config = {
            "model_map": self.model_map[model_name],
            "batch_size": batch_size,
            "device": "cuda",
        }



import hydra
from omegaconf import DictConfig
from helical.models.transcriptformer.model import TranscriptFormer
from helical.models.transcriptformer.transcriptformer_config import (
    TranscriptFormerConfig,
)
from datasets import load_dataset
from helical.utils import get_anndata_from_hf_dataset
import anndata as ad


@hydra.main(
    version_base=None, config_path="configs", config_name="transcriptformer_config"
)
def run(cfg: DictConfig):
    configurer = TranscriptFormerConfig(**cfg)
    model = TranscriptFormer(configurer)

    # # either load via huggingface
    # hf_dataset = load_dataset("helical-ai/yolksac_human",split="train[:5%]", trust_remote_code=True, download_mode="reuse_cache_if_exists")
    # ann_data = get_anndata_from_hf_dataset(hf_dataset)

    # or load directly
    ann_data = ad.read_h5ad("./yolksac_human.h5ad")

    dataset = model.process_data([ann_data[:10]])
    embeddings = model.get_embeddings(dataset)
    print(embeddings)
    log_likelihoods = model.get_output_adata().uns["llh"]
    print(log_likelihoods)


if __name__ == "__main__":
    run()



import torch
from abc import ABC
from abc import abstractmethod


class HelicalBaseFineTuningHead(torch.nn.Module, ABC):
    """Helical Fine-Tuning Head Class which serves as the base class for all fine-tuning heads in the helical package.
    Each new fine-tuning head should be a subclass of this class.
    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def forward():
        pass

    @abstractmethod
    def set_dim_size():
        pass


class ClassificationHead(HelicalBaseFineTuningHead):
    """Classification Head for fine-tuning a Helical foundation model.

    Parameters
    ----------
    num_classes : int
        The number of classes to predict.
    dropout : float, optional, default=0.02
        The dropout rate to apply to the input tensor before the linear layer.

    Methods
    -------
    forward(inputs: torch.Tensor) -> torch.Tensor
        The forward method of the classification head.

    """

    def __init__(self, num_classes: int, dropout: float = 0.02):
        super().__init__()
        self.output_size = num_classes
        self.dropout = torch.nn.Dropout(p=dropout)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Forward method of the classification head.

        Parameters
        ----------
        inputs : torch.Tensor
            The input tensor to the classification head.

        Returns
        -------
        torch.Tensor
            The output tensor of the classification head.
        """
        drop = self.dropout(inputs)
        output = self.linear(drop)
        return output

    def set_dim_size(self, dim_size: int) -> None:
        """Set the dimension size of the input tensor.

        Parameters
        ----------
        dim_size : int
            The dimension size of the input tensor.
        """
        self.linear = torch.nn.Linear(dim_size, self.output_size)


class RegressionHead(HelicalBaseFineTuningHead):
    """Regression Head for fine-tuning a Helical foundation model.

    Parameters
    ----------
    num_outputs : int
        The number of outputs to predict.
    dropout : float, optional, default=0.02
        The dropout rate to apply to the input tensor before the linear layer.

    Methods
    -------
    forward(inputs: torch.Tensor) -> torch.Tensor
        The forward method of the regression head.

    """

    def __init__(self, num_outputs: int, dropout: float = 0.02):
        super().__init__()
        self.output_size = num_outputs
        self.dropout = torch.nn.Dropout(p=dropout)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Forward method of the classification head.

        Parameters
        ----------
        inputs : torch.Tensor
            The input tensor to the regression head.

        Returns
        -------
        torch.Tensor
            The output tensor of the regression head.
        """
        drop = self.dropout(inputs)
        output = self.linear(drop)
        return output

    def set_dim_size(self, dim_size: int) -> None:
        """Set the dimension size of the input tensor.

        Parameters
        ----------
        dim_size : int
            The dimension size of the input tensor.
        """
        self.linear = torch.nn.Linear(dim_size, self.output_size)



"""
Helper functions for loading pretrained gene embeddings.
"""

from pathlib import Path
from typing import Dict, Tuple
import torch
from scanpy import AnnData
import logging
from helical.models.uce.uce_config import SPECIES_GENE_EMBEDDINGS

LOGGER = logging.getLogger(__name__)


def get_gene_embedding_paths(embedding_path: Path) -> Dict[str, Dict[str, Path]]:
    """
    Get the paths to the gene embeddings for all species and models by prepending the embedding path to the file names.

    :param embedding_path: The path to the directory containing the gene embeddings.
    :return: A dictionary mapping model names to dictionaries mapping species names to the paths of the gene embeddings.
    """
    res = {}
    for model, species_dict in SPECIES_GENE_EMBEDDINGS.items():
        res[model] = {}
        for species, file in species_dict.items():
            res[model][species] = embedding_path / file
    return res


# TODO Add new function to add embeddings
# extra_species = pd.read_csv("./UCE/model_files/new_species_protein_embeddings.csv").set_index("species").to_dict()["path"]
# MODEL_TO_SPECIES_TO_GENE_EMBEDDING_PATH["ESM2"].update(extra_species) # adds new species


def load_gene_embeddings_adata(
    adata: AnnData, species: list, embedding_model: str, embeddings_path: Path
) -> Tuple[AnnData, Dict[str, list[str]]]:
    """
    Loads gene embeddings for all the species/genes in the provided data.

    :param adata: An AnnData object containing gene expression data for cells.
    :param species: Species corresponding to this adata
    :param embedding_model: The gene embedding model whose embeddings will be loaded.
    :param embedding_model: The embedding paths for the model
    :return: A tuple containing:
               - A subset of the data only containing the gene expression for genes with embeddings in all species.
               - A dictionary mapping species name to the names of genes.
    """
    # Get species names
    species_names = species
    species_names_set = set(species_names)

    # Get embedding paths for the model
    embedding_paths = get_gene_embedding_paths(embeddings_path)
    species_to_gene_embedding_path = embedding_paths[embedding_model]
    available_species = set(species_to_gene_embedding_path)

    # Ensure embeddings are available for all species
    if not (species_names_set <= available_species):
        LOGGER.error(f"Missing gene embeddings here: {embeddings_path}")
        raise ValueError(
            f"The following species do not have gene embeddings: {species_names_set - available_species}"
        )
    # Load gene embeddings for desired species (and convert gene symbols to lower case)
    species_to_gene_symbol_to_embedding = {
        species: {
            gene_symbol.lower(): gene_embedding
            for gene_symbol, gene_embedding in torch.load(
                species_to_gene_embedding_path[species]
            ).items()
        }
        for species in species_names
    }

    LOGGER.info(
        f"Finished loading gene embeddings for '{', '.join(map(str, species_names_set))}' from {embeddings_path} for model '{embedding_model}'."
    )

    # Determine which genes to include based on gene expression and embedding availability
    genes_with_embeddings = set.intersection(
        *[
            set(gene_symbol_to_embedding)
            for gene_symbol_to_embedding in species_to_gene_symbol_to_embedding.values()
        ]
    )
    genes_to_use = {
        gene for gene in adata.var_names if gene.lower() in genes_with_embeddings
    }

    # Subset data to only use genes with embeddings
    filtered_adata = adata[:, adata.var_names.isin(genes_to_use)]
    filtered = adata.var_names.shape[0] - filtered_adata.var_names.shape[0]
    LOGGER.info(
        f"Filtered out {filtered} genes to a total of {filtered_adata.var_names.shape[0]} genes with embeddings."
    )

    if filtered_adata.var_names.shape[0] == 0:
        message = "No matching genes found between input data and UCE gene embedding vocabulary. Please check the gene names in .var of the anndata input object."
        LOGGER.error(message)
        raise ValueError(message)

    # Load gene symbols for desired species for later use with indexes
    species_to_all_gene_symbols = {
        species: [
            gene_symbol.lower()
            for gene_symbol, _ in torch.load(
                species_to_gene_embedding_path[species]
            ).items()
        ]
        for species in species_names
    }
    return filtered_adata, species_to_all_gene_symbols



import logging
from helical.constants.enums import LoggingType, LoggingLevel
import os


class Logger:
    def __init__(self, log_type: LoggingType, level: LoggingLevel):

        self.log_filename = "debug.log"
        format = "%(asctime)s.%(msecs)03d %(name)s %(levelname)s %(message)s"
        datetime = "%Y-%m-%d, %H:%M:%S"
        level = level.value

        if log_type == LoggingType.FILE_AND_CONSOLE:
            logging.basicConfig(
                format=format,
                datefmt=datetime,
                level=level,
                handlers=[
                    logging.FileHandler(self.log_filename),
                    logging.StreamHandler(),
                ],
            )

        elif log_type == LoggingType.CONSOLE:
            logging.basicConfig(
                format=format,
                datefmt=datetime,
                level=level,
                handlers=[
                    logging.StreamHandler(),
                ],
            )

        elif log_type == LoggingType.FILE:
            logging.basicConfig(
                format=format,
                datefmt=datetime,
                level=level,
                handlers=[
                    logging.FileHandler(self.log_filename),
                ],
            )

        else:
            logging.disable = True



import logging
import os
import random

import anndata
import numpy as np
import scanpy as sc
import torch
from scipy.sparse import csc_matrix, csr_matrix
from torch import tensor
from torch.utils.data import Dataset
from helical.utils.mapping import map_gene_symbols_to_ensembl_ids
from helical.models.transcriptformer.data.dataclasses import BatchData
from helical.models.transcriptformer.tokenizer.tokenizer import (
    BatchGeneTokenizer,
    BatchObsTokenizer,
)


def load_data(file_path):
    """Load H5AD file."""
    try:
        adata = sc.read_h5ad(file_path)
        return adata, True
    except Exception as e:
        logging.error(f"Failed to read file {file_path}: {e}")
        return None, False


def load_gene_features(adata, gene_col_name):
    """Load ensembl ids from adata object."""
    if gene_col_name == "index":
        adata.var["index"] = adata.var_names
    elif gene_col_name not in adata.var.columns:
        message = f"Gene column '{gene_col_name}' not found in adata.var.columns. Available columns: {adata.var.columns}. Modify config accordingly."
        logging.error(message)
        raise ValueError(message)
    adata = map_gene_symbols_to_ensembl_ids(adata, gene_names=gene_col_name)
    gene_names = np.array(list(adata.var["ensembl_id"].values))
    return gene_names, True


def apply_filters(
    X,
    obs,
    gene_names,
    file_path,
    filter_to_vocab,
    vocab,  # gene  vocab
    filter_outliers,
    min_expressed_genes,
):
    """Apply filters to the data."""
    if filter_to_vocab:
        filter_idx = [i for i, name in enumerate(gene_names) if name in vocab]
        X = X[:, filter_idx]
        gene_names = gene_names[filter_idx]
        if X.shape[1] == 0:
            logging.warning(f"Warning: Filtered all genes from {file_path}")
            logging.warning(f"Available genes: {len(gene_names)}")
            logging.warning(f"Number of non-zero genes: {np.sum(X > 0, axis=1).mean()}")
            return None, None, None

    if filter_outliers > 0:
        expr_counts = X.sum(axis=1)
        count_std = np.std(expr_counts)
        count_mean = np.mean(expr_counts)
        filter_idx = (expr_counts > count_mean - count_std * filter_outliers) & (
            expr_counts < count_mean + count_std * filter_outliers
        )
        X = X[filter_idx]
        obs = obs.iloc[filter_idx]

    if min_expressed_genes > 0:
        filter_idx = (X > 0).sum(axis=1) >= min_expressed_genes
        X = X[filter_idx]
        obs = obs.iloc[filter_idx]

    return X, obs, gene_names


def process_batch(
    x_batch,
    obs_batch,
    gene_names,
    gene_tokenizer,
    aux_tokenizer,
    sort_genes,
    randomize_order,
    max_len,
    pad_zeros,
    pad_token,
    gene_vocab,
    normalize_to_scale,
    clip_counts,
    aux_vocab,
):
    """Process a batch of data, including sorting, tokenization, and normalization."""
    x_batch = tensor(x_batch, dtype=torch.float32)

    # Sort genes or randomize order
    if sort_genes:
        ids_batch = torch.argsort(x_batch, dim=1, descending=True)
    else:
        ids_batch = torch.zeros_like(x_batch, dtype=torch.long)
        for i, sample in enumerate(x_batch):
            non_zero_indices = torch.nonzero(sample, as_tuple=True)[0]
            zero_indices = torch.nonzero(sample == 0, as_tuple=True)[0]
            if randomize_order:
                non_zero_indices = non_zero_indices[
                    torch.randperm(len(non_zero_indices))
                ]
                zero_indices = zero_indices[torch.randperm(len(zero_indices))]
            sample_ids = torch.cat([non_zero_indices, zero_indices])
            ids_batch[i] = sample_ids

    # Limit to max_len and gather counts
    if ids_batch.shape[1] > max_len:
        ids_batch = ids_batch[:, :max_len]

    counts_batch = torch.gather(x_batch, 1, ids_batch)

    # Tokenize gene names
    gene_names_batch = gene_names[ids_batch.numpy()]
    gene_tokens_batch = gene_tokenizer(gene_names_batch)

    # Apply padding and normalization
    if pad_zeros:
        gene_tokens_batch = gene_tokens_batch.masked_fill(
            counts_batch == 0, gene_vocab[pad_token]
        )

    # Pad ids_batch to max_len
    tok_bz, tok_sq = gene_tokens_batch.shape
    if tok_sq < max_len:
        padding = torch.full(
            (tok_bz, max_len - tok_sq),
            gene_vocab[pad_token],
            dtype=gene_tokens_batch.dtype,
        )
        gene_tokens_batch = torch.cat([gene_tokens_batch, padding], dim=1)
        gene_names_batch = np.hstack(
            [
                gene_names_batch,
                np.full((tok_bz, max_len - tok_sq), pad_token),
            ]
        )

        counts_batch = torch.cat(
            [counts_batch, torch.zeros_like(padding, dtype=counts_batch.dtype)], dim=1
        )

    # Normalize to scale if specified
    if normalize_to_scale is not None and normalize_to_scale > 0:
        row_sums = counts_batch.sum(dim=1, keepdim=True)
        counts_batch = counts_batch / row_sums * normalize_to_scale

    # Clip counts if specified
    if clip_counts is not None:
        counts_batch = counts_batch.clamp(min=0, max=clip_counts)

    # Prepare result dictionary
    result = {
        "gene_counts": counts_batch,
        "gene_token_indices": gene_tokens_batch,
    }

    # Add auxiliary and tokens if specified
    if aux_vocab is not None:
        aux_tokens_batch = torch.stack(
            [aux_tokenizer(obs) for _, obs in obs_batch.iterrows()]
        )
        result["aux_token_indices"] = aux_tokens_batch

    return result


class AnnDataset(Dataset):
    def __init__(
        self,
        files_list: list[str] | list[anndata.AnnData],
        gene_vocab: dict[str, str],
        data_dir: str = None,
        aux_vocab: dict[str, dict[str, str]] = None,
        max_len: int = 2000,
        normalize_to_scale: bool = None,
        sort_genes: bool = False,
        randomize_order: bool = False,
        pad_zeros: bool = True,
        gene_col_name: str = "feature_id",
        filter_to_vocab: bool = True,
        filter_outliers: float = 0.0,
        min_expressed_genes: int = 200,
        seed: int = 0,
        pad_token: str = "[PAD]",
        clip_counts: float = 1e10,
        inference: bool = False,
        obs_keys: list[str] = None,
    ):
        super().__init__()
        self.data_dir = data_dir
        self.files_list = files_list
        self.gene_vocab = gene_vocab
        self.aux_vocab = aux_vocab
        self.max_len = max_len
        self.normalize_to_scale = normalize_to_scale
        self.sort_genes = sort_genes
        self.randomize_order = randomize_order
        self.pad_zeros = pad_zeros
        self.gene_col_name = gene_col_name
        self.filter_to_vocab = filter_to_vocab
        self.filter_outliers = filter_outliers
        self.min_expressed_genes = min_expressed_genes
        self.seed = seed
        self.pad_token = pad_token
        self.clip_counts = clip_counts
        self.inference = inference
        self.obs_keys = obs_keys

        self.gene_tokenizer = BatchGeneTokenizer(gene_vocab)
        if aux_vocab is not None:
            self.aux_tokenizer = BatchObsTokenizer(aux_vocab)

        random.seed(self.seed)

        logging.info("Loading and processing all data")
        self.data = self.load_and_process_all_data()

    def _get_batch_from_file(self, file: str | anndata.AnnData) -> BatchData | None:
        if isinstance(file, str):
            file_path = (
                file if self.data_dir is None else os.path.join(self.data_dir, file)
            )
            adata, success = load_data(file_path)
        elif isinstance(file, anndata.AnnData):
            adata = file
            success = True
            file_path = None
        else:
            raise ValueError(f"Invalid file type: {type(file)}")

        if not success:
            logging.error(
                f"Failed to load data from {file_path if file_path else 'provided AnnData object'}"
            )
            return None

        gene_names, success = load_gene_features(adata, self.gene_col_name)
        if not success:
            logging.error(
                f"Failed to load gene features from {file_path if file_path else 'provided AnnData object'}"
            )
            return None

        X = (
            adata.X.toarray()
            if isinstance(adata.X, csr_matrix | csc_matrix)
            else adata.X
        )
        obs = adata.obs

        if not hasattr(obs, "assay"):
            logging.warning(
                f"'assay' column not found in {file_path if file_path else 'provided AnnData object'}. Adding 'unknown' as default."
            )
            obs["assay"] = "unknown"

        vocab = self.gene_vocab
        X, obs, gene_names = apply_filters(
            X,
            obs,
            gene_names,
            file_path,
            self.filter_to_vocab,
            vocab,
            self.filter_outliers,
            self.min_expressed_genes,
        )
        if X is None:
            logging.warning(
                f"Data was filtered out completely for {file_path if file_path else 'provided AnnData object'}"
            )
            return None

        batch = process_batch(
            X,
            obs,
            gene_names,
            self.gene_tokenizer,
            getattr(self, "aux_tokenizer", None),
            self.sort_genes,
            self.randomize_order,
            self.max_len,
            self.pad_zeros,
            self.pad_token,
            self.gene_vocab,
            self.normalize_to_scale,
            self.clip_counts,
            self.aux_vocab,
        )
        batch["file_path"] = (
            np.array([file_path] * X.shape[0]) if file_path is not None else None
        )

        if self.obs_keys is not None:
            obs_data = {}
            if "all" in self.obs_keys:
                # Keep all columns from obs
                self.obs_keys = obs.columns
                for col in obs.columns:
                    obs_data[col] = np.array(obs[col].tolist())[:, None]
            else:
                # Keep only specified columns
                for col in self.obs_keys:
                    obs_data[col] = np.array(obs[col].tolist())[:, None]
            batch["obs"] = obs_data

        return BatchData(**batch)

    def load_and_process_all_data(self):
        all_data = []
        for i, file in enumerate(self.files_list):
            logging.info(f"Processing validation file {i+1} of {len(self.files_list)}")
            file_batch = self._get_batch_from_file(file)
            if file_batch is None:
                continue

            all_data.append(file_batch)

        # Add check for empty all_data list
        if not all_data:
            raise ValueError(
                "No valid data was loaded from any files. "
                "Check if files exist and contain valid data after filtering."
            )

        concatenated_batch = BatchData(
            gene_counts=torch.concat([batch.gene_counts for batch in all_data]),
            gene_token_indices=torch.concat(
                [batch.gene_token_indices for batch in all_data]
            ),
            file_path=None,
            aux_token_indices=(
                torch.concat([batch.aux_token_indices for batch in all_data])
                if all_data[0].aux_token_indices is not None
                else None
            ),
            obs=(
                {
                    col: np.vstack([batch.obs[col] for batch in all_data])
                    for col in self.obs_keys
                }
                if self.obs_keys is not None
                else None
            ),
        )

        return concatenated_batch

    def __len__(self):
        return len(self.data.gene_counts)

    def __getitem__(self, idx):
        data_dict = {}
        for key, value in self.data.__dict__.items():
            if value is None:
                data_dict[key] = None
            elif isinstance(value, dict):
                data_dict[key] = {k: v[idx] for k, v in value.items()}
            else:
                data_dict[key] = value[idx]
        return BatchData(**data_dict)

    @staticmethod
    def collate_fn(batch: BatchData | list[BatchData]) -> BatchData:
        if isinstance(batch, BatchData):
            return batch

        collated_batch = BatchData(
            gene_counts=torch.stack([item.gene_counts for item in batch]),
            gene_token_indices=torch.stack([item.gene_token_indices for item in batch]),
            file_path=None,
            aux_token_indices=(
                torch.stack([item.aux_token_indices for item in batch])
                if batch[0].aux_token_indices is not None
                else None
            ),
            obs=(
                {
                    col: np.vstack([item.obs[col] for item in batch])
                    for col in batch[0].obs.keys()
                }
                if batch[0].obs is not None
                else None
            ),
        )
        return collated_batch



from dataclasses import dataclass
from typing import Dict, List, Mapping, Optional, Tuple
import torch
from .binning import binning


@dataclass
class DataCollator:
    """
    Data collator for the mask value learning task. It pads the sequences to
    the maximum length in the batch and masks the gene expression values.

    Args:
        do_padding (:obj:`bool`): whether to pad the sequences to the max length.
        pad_token_id (:obj:`int`, optional): the token id to use for padding.
            This is required if do_padding is True.
        pad_value (:obj:`int`): the value to use for padding the expression
            values to the max length.
        do_mlm (:obj:`bool`): whether to do masking with MLM.
        do_binning (:obj:`bool`): whether to bin the expression values.
        mlm_probability (:obj:`float`): the probability of masking with MLM.
        mask_value (:obj:`int`): the value to fill at the expression postions
            that are masked.
        max_length (:obj:`int`, optional): the maximum length of the sequences.
            This is required if do_padding is True.
        sampling (:obj:`bool`): whether to do sampling instead of truncation if
            length > max_length.
        keep_first_n_tokens (:obj:`int`): the number of tokens in the beginning
            of the sequence to keep unchanged from sampling. This is useful when
            special tokens have been added to the beginning of the sequence.
            Default to 1.
    """

    do_padding: bool = True
    pad_token_id: Optional[int] = None
    pad_value: int = 0
    do_mlm: bool = True
    do_binning: bool = True
    mlm_probability: float = 0.15
    mask_value: int = -1
    max_length: Optional[int] = None
    sampling: bool = True
    keep_first_n_tokens: int = 1

    def __post_init__(self):
        if self.do_padding:
            if self.pad_token_id is None:
                raise ValueError("`pad_token_id` is required if `do_padding`.")
            if self.max_length is None:
                raise ValueError("`max_length` is required if `do_padding`.")

        if self.mlm_probability <= 0 or self.mlm_probability >= 1:
            raise ValueError("`mlm_probability` must be between 0 and 1.")

        if self.keep_first_n_tokens < 0 or self.keep_first_n_tokens > self.max_length:
            raise ValueError(
                "`keep_first_n_tokens` must be between 0 and `max_length` "
                f"({self.max_length})."
            )

    def __call__(
        self, examples: List[Dict[str, torch.Tensor]]
    ) -> Dict[str, torch.Tensor]:
        """
        Each example is like:
            {'id': tensor(184117),
            'genes': tensor([36572, 17868, ..., 17072]),
            'expressions': tensor([ 0.,  2., ..., 18.])}
        """
        if not isinstance(examples[0], Mapping):
            return NotImplementedError

        device = examples[0]["genes"].device

        max_ori_len = max(len(example["genes"]) for example in examples)
        _max_length = self.max_length if max_ori_len >= self.max_length else max_ori_len

        # pad and truncate
        padded_genes = []
        padded_expressions = []
        for i in range(len(examples)):
            genes = examples[i]["genes"]
            expressions = examples[i]["expressions"]
            if self.do_binning:
                expressions[self.keep_first_n_tokens :] = binning(
                    row=expressions[self.keep_first_n_tokens :],
                    n_bins=51,
                )
            genes, expressions = self._sample_or_truncate_plus_pad(
                genes, expressions, _max_length
            )  # torch tensors of length _max_length
            padded_genes.append(genes)
            padded_expressions.append(expressions)

        padded_genes = torch.stack(padded_genes, dim=0).to(device)
        padded_expressions = torch.stack(padded_expressions, dim=0).to(device)

        data_dict = {
            "gene": padded_genes,
            "expr": padded_expressions,
        }

        # mask
        if self.do_mlm:
            masked_expressions = self._mask(padded_expressions)
        else:
            masked_expressions = padded_expressions
        data_dict["masked_expr"] = masked_expressions

        return data_dict

    def _mask(self, expressions: torch.Tensor) -> torch.Tensor:
        """
        Mask the expression values with MLM.
        """
        device = expressions.device
        shape = expressions.shape

        probability_matrix = torch.full(shape, self.mlm_probability)
        # set padded postion probability to 0
        probability_matrix[expressions.eq(self.pad_value)] = 0
        if self.keep_first_n_tokens > 0:
            probability_matrix[:, : self.keep_first_n_tokens] = 0

        mask = torch.bernoulli(probability_matrix).bool()
        mask = mask.to(device)

        masked_expressions = expressions.masked_fill(mask, self.mask_value)
        return masked_expressions

    def _sample_or_truncate_plus_pad(
        self,
        genes: torch.LongTensor,
        expressions: torch.Tensor,
        max_length: int,
    ) -> Tuple[torch.LongTensor, torch.Tensor]:
        assert len(genes) == len(expressions)
        if len(genes) == max_length:
            return genes, expressions
        if len(genes) > max_length:  # sample or truncate
            if self.sampling:
                return self._sample(genes, expressions, max_length)
            else:
                return genes[:max_length], expressions[:max_length]
        else:  # pad
            return self._pad(genes, expressions, max_length)

    def _sample(
        self,
        genes: torch.LongTensor,
        expressions: torch.Tensor,
        max_length: int,
    ) -> Tuple[torch.LongTensor, torch.Tensor]:
        # NOTE: the fastest way to sample in torch has been benchmarked here
        # https://discuss.pytorch.org/t/torch-equivalent-of-numpy-random-choice/16146/19
        # it shows the randperm on gpu is the fastest.
        # NOTE: also, the current implementation permute the orders of the genes
        # and expressions, although it is probably a nice argmentation.
        device = genes.device
        if self.keep_first_n_tokens == 0:
            indices = torch.randperm(len(genes), device=device)[:max_length]
            return genes[indices], expressions[indices]

        # keep the first n tokens unchanged
        _n = self.keep_first_n_tokens
        indices = torch.randperm(len(genes) - _n, device=device)[: max_length - _n]
        indices = torch.cat([torch.arange(_n), indices + _n], dim=0)
        return genes[indices], expressions[indices]

    def _pad(
        self,
        genes: torch.LongTensor,
        expressions: torch.Tensor,
        max_length: int,
    ):
        device = genes.device
        genes = torch.cat(
            [
                genes,
                torch.full(
                    (max_length - len(genes),),
                    self.pad_token_id,
                    dtype=genes.dtype,
                    device=device,
                ),
            ]
        )
        expressions = torch.cat(
            [
                expressions,
                torch.full(
                    (max_length - len(expressions),),
                    self.pad_value,
                    dtype=expressions.dtype,
                    device=device,
                ),
            ]
        )
        return genes, expressions



import os
import logging

logging.captureWarnings(True)

for handler in logging.root.handlers[:]:
    logging.root.removeHandler(handler)

logger = logging.getLogger()
logger.setLevel(logging.INFO)

handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
handler.setFormatter(formatter)

logger.addHandler(handler)

os.environ["TRANSFORMERS_VERBOSITY"] = "error"



import copy

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch.nn.attention.flex_attention import flex_attention


def clones(module, N):
    """Produce N identical layers."""
    return torch.nn.ModuleList([copy.deepcopy(module) for _ in range(N)])


def mean_embeddings(embeddings, mask=None, include_masked=True):
    if mask is None:
        return embeddings.mean(dim=1)
    elif include_masked:
        return (embeddings * mask.unsqueeze(-1)).mean(dim=1)
    else:
        return (embeddings * mask.unsqueeze(-1)).sum(dim=1) / mask.sum(dim=1)


class MultiHeadSelfFlexAttn(nn.Module):
    """
    Multi-Head Self-Attention with Flex Attention mechanism.

    This module implements a multi-head self-attention layer using flex attention,
    which allows for custom score modifications and block masking.

    Args:
        d_model (int): The dimension of the model (input and output)
        nheads (int): The number of attention heads
        bias (bool, optional): Whether to include bias in linear layers. Defaults to False.

    Attributes
    ----------
        d_k (int): The dimension of keys/queries in each head
        h (int): The number of attention heads
        linears (nn.ModuleList): List of linear transformations for Q, K, V, and output
        attn (None): Placeholder for attention scores (not used in this implementation)

    """

    def __init__(self, d_model: int, nheads: int, bias: bool = False):
        super().__init__()
        assert d_model % nheads == 0, "d_model must be divisible by number of heads"

        # We assume d_v always equals d_k
        self.d_k = d_model // nheads
        self.h = nheads
        self.linears = clones(nn.Linear(d_model, d_model, bias=bias), 4)
        self.self_attn = flex_attention

    def forward(self, inp, score_mod=None, block_mask=None, **kwargs):
        """Forward pass of the MultiHeadSelfFlexAttn."""
        batch_size = inp.size(0)

        # 1) Do all the linear projections in batch from d_model => h x d_k
        q, k, v = (
            layer(x).view(batch_size, -1, self.h, self.d_k).transpose(1, 2)
            for layer, x in zip(self.linears, [inp] * 3, strict=False)
        )

        # 2) Apply attention on all the projected vectors in batch.
        o = self.self_attn(
            q,
            k,
            v,
            score_mod=score_mod,
            block_mask=block_mask,
            kernel_options={
                "BLOCK_M": 32,
                "BLOCK_N": 32,
                "BLOCK_M1": 32,
                "BLOCK_N1": 32,
                "BLOCK_M2": 32,
                "BLOCK_N2": 32,
            },
        )

        # 3) "Concat" using a view and apply a final linear.
        o = o.transpose(1, 2).contiguous().view(batch_size, -1, self.h * self.d_k)

        return self.linears[-1](o)


class FlexAttnTransformerLayer(nn.Module):
    """
    Flex Attention Transformer Layer

    This layer implements a transformer block with flex attention mechanism.
    It includes self-attention with customizable score modifications and block masking,
    followed by a feed-forward network.

    Args:
        d_model (int): The dimension of the model (input and output)
        nhead (int): The number of attention heads
        dim_fw (int, optional): The dimension of the feed-forward network. Defaults to 2048.
        dropout (float, optional): Dropout rate. Defaults to 0.0.
        fw_bias (bool, optional): Whether to include bias in feed-forward layers. Defaults to False.
        attn_bias (bool, optional): Whether to include bias in attention layers. Defaults to False.
        activation (str, optional): Activation function to use. Can be 'relu', 'gelu', or 'silu'. Defaults to 'gelu'.

    Attributes
    ----------
        linear1 (nn.Linear): First linear layer of the feed-forward network
        linear2 (nn.Linear): Second linear layer of the feed-forward network
        self_attn (MultiHeadSelfFlexAttn): Multi-head self-attention layer with flex attention
        norm1 (nn.LayerNorm): Layer normalization for attention output
        norm2 (nn.LayerNorm): Layer normalization for feed-forward output
        dropout (nn.Dropout): Dropout layer
        activation (nn.Module): Activation function

    """

    def __init__(
        self,
        d_model: int,
        nhead: int,
        dim_fw: int = 2048,
        dropout: float = 0.0,
        fw_bias: bool = False,
        attn_bias: bool = False,
        activation: str = "gelu",
    ):
        super().__init__()

        self.linear1 = nn.Linear(d_model, dim_fw, bias=fw_bias)
        self.linear2 = nn.Linear(dim_fw, d_model, bias=fw_bias)
        self.self_attn = MultiHeadSelfFlexAttn(d_model, nhead, bias=attn_bias)
        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)

        if activation == "relu":
            self.activation = nn.ReLU()
        elif activation == "gelu":
            self.activation = nn.GELU()
        elif activation == "silu":
            self.activation = nn.SiLU()
        else:
            raise ValueError(f"Activation {activation} not supported")

    def forward(self, x, score_mod=None, block_mask=None):
        """Forward pass of the FlexAttnTransformerLayer."""
        x = self.norm1(x)  # pre-norm is the new norm
        x = self.self_attn(x, score_mod=score_mod, block_mask=block_mask) + x
        x = self.norm2(x)
        x = self.linear2(self.dropout(self.activation(self.linear1(x)))) + x
        return x


class TranscriptEncoder(nn.Module):
    """
    TranscriptEncoder module for encoding transcript sequences.

    This module applies a series of FlexAttnTransformerLayers to the input sequence,
    incorporating count-based attention bias and optional masking.

    Args:
        embed_dim (int): The dimension of the input embeddings.
        num_head (int): The number of attention heads in each transformer layer.
        nlayers (int): The number of transformer layers.
        model_dim (int, optional): The dimension of the feed-forward network in each layer. Defaults to 2048.
        dropout (float, optional): Dropout rate. Defaults to 0.0.
        activation (str, optional): Activation function to use. Can be 'relu', 'gelu', or 'silu'. Defaults to 'gelu'.
        attn_bias (bool, optional): Whether to include bias in attention layers. Defaults to False.
        fw_bias (bool, optional): Whether to include bias in feed-forward layers. Defaults to False.
        eps (float, optional): Small constant for numerical stability. Defaults to 1e-6.
        softcap (int, optional): If > 0, applies a soft cap to attention scores. Defaults to 0.

    Attributes
    ----------
        nlayers (int): The number of transformer layers.
        nheads (int): The number of attention heads in each layer.
        eps (float): Small constant for numerical stability.
        score_mod_factory (function): Factory function for creating score modifiers.
        mask_mod_factory (function): Factory function for creating mask modifiers.
        encoder_layers (nn.ModuleList): List of FlexAttnTransformerLayer modules.

    """

    def __init__(
        self,
        embed_dim: int,
        num_head: int,
        nlayers: int,
        model_dim: int = 2048,
        dropout: float = 0.0,
        activation: str = "gelu",
        attn_bias: bool = False,
        fw_bias: bool = False,
    ):
        super().__init__()
        self.nlayers = nlayers
        self.nheads = num_head

        self.encoder_layers = nn.ModuleList(
            [
                FlexAttnTransformerLayer(
                    d_model=embed_dim,
                    nhead=num_head,
                    dim_fw=model_dim,
                    dropout=dropout,
                    fw_bias=fw_bias,
                    attn_bias=attn_bias,
                    activation=activation,
                )
                for _ in range(nlayers)
            ]
        )

    def forward(self, x, score_mod, block_mask):
        """
        Forward pass of the TranscriptEncoder.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, seq_len, embed_dim).
            score_mod (torch.Tensor): Score modifier tensor for attention.
            block_mask (torch.Tensor): Block mask tensor for attention.

        Returns
        -------
            torch.Tensor: Output tensor of shape (batch_size, seq_len, embed_dim).
        """
        for layer in self.encoder_layers:
            x = layer(x, score_mod=score_mod, block_mask=block_mask)
        return x


class MaskedSoftmax(nn.Module):
    def __init__(self, dim: int, softcap: int = 0):
        super().__init__()
        self.dim = dim
        self.softcap = softcap

    def forward(self, x: Tensor, m: Tensor) -> Tensor:
        """
        Forward pass.

        x: (batch, seq_len, d_model)
        m: (batch, seq_len, d_model)

        m is a additive matrix that is added to x before applying softmax
        masked values are set to -inf
        """
        if self.softcap > 0:
            x = x / self.softcap
            x = torch.tanh(x)
            x = x * self.softcap
        return F.softmax(x + m, dim=self.dim)


class CountDecoderHead(nn.Module):
    """
    A decoder head that predicts gene expression counts.

    This module takes gene embeddings and optionally gene token indices as input,
    and outputs predicted expression values using a specified link function.
    """

    def __init__(
        self,
        model_dim: int,  # Dimension of the input embeddings
        link_func: str,  # Type of activation function to use
        eps: float = 0,  # Small epsilon value for numerical stability
        dropout: float = 0.1,  # Dropout rate for regularization
        use_layer_norm: bool = True,  # Whether to use layer normalization
        gene_bias_size: int = 0,  # Size of gene-specific bias embedding
        softcap: int = 0,  # Cap value for softmax activation
    ):
        super().__init__()
        self.seq_dim = model_dim
        self.link_func = link_func  # Store the name for reference
        self.eps = eps
        self.gene_bias_size = gene_bias_size
        self.softcap = softcap

        # Optional gene-specific bias term
        if gene_bias_size > 0:
            self.gene_embedding = nn.Embedding(gene_bias_size, 1)

        # Main MLP for processing gene embeddings
        self.mlp = MLP(
            model_dim,  # Input dimension
            model_dim // 4,  # Hidden dimension (1/4 of input)
            1,  # Output dimension (scalar count)
            num_layers=2,  # Two-layer MLP
            dropout=dropout,
            use_layer_norm=use_layer_norm,
        )

        # Initialize the appropriate activation function based on link_func
        if link_func == "sigmoid":
            self.link_func = nn.Sigmoid()
        elif link_func == "softplus":
            self.link_func = nn.Softplus()
        elif link_func == "relu":
            self.link_func = nn.ReLU()
        elif link_func == "linear":
            self.link_func = nn.Identity()
        elif link_func == "exp":
            self.link_func = torch.exp
        elif link_func == "log":
            self.link_func = torch.log
        elif link_func == "tanh":
            self.link_func = nn.Tanh()
        elif link_func == "softmax":
            # Special case: masked softmax with optional softcap
            self.link_func = MaskedSoftmax(dim=1, softcap=softcap)
        else:
            raise ValueError(f"Invalid link function: {link_func}")

    def forward(
        self,
        gene_output: Tensor = None,  # Gene embeddings from transformer
        gene_tokens: Tensor = None,  # Gene token indices (optional)
        mask: Tensor = None,  # Mask for softmax (optional)
    ) -> Tensor:
        """
        Forward pass to predict gene expression counts.

        Args:
            gene_output: Embeddings from transformer, shape [batch, seq_len, model_dim]
            gene_tokens: Gene token indices, shape [batch, seq_len]
            mask: Mask tensor for softmax, shape [batch, seq_len]

        Returns
        -------
            Predicted gene expression values, shape [batch, seq_len]
        """
        # Initialize logits
        logit = 0

        # Add gene-specific bias if enabled
        if self.gene_bias_size > 0:
            gene_emb = self.gene_embedding(gene_tokens)  # [batch, seq_len, 1]
            logit = logit + gene_emb

        # Process gene embeddings through MLP
        logit = logit + self.mlp(gene_output)  # [batch, seq_len, 1]

        # Remove singleton dimension if needed
        logit = logit.squeeze(-1) if logit.dim() > 2 else logit  # [batch, seq_len]

        # Apply the link function (with mask if using MaskedSoftmax)
        if mask is not None and isinstance(self.link_func, MaskedSoftmax):
            output = self.link_func(logit, mask)
        else:
            output = self.link_func(logit)

        return output

    def init_weights(self):
        """Initialize the gene embedding weights to zero if used."""
        if self.gene_bias_size > 0:
            self.gene_embedding.weight.data.zero_()


class PretrainedEmbeddings(nn.Module):
    def __init__(
        self,
        embedding_matrix,
        output_dim,
        freeze=True,
        normalize=True,
        mlp_hidden_dim=None,
        dropout=0.1,
    ):
        super().__init__()
        self.embedding_dim = embedding_matrix.shape[1]

        # Check embedding_matrix for NaNs
        if torch.isnan(embedding_matrix).any():
            raise ValueError("The embedding matrix contains NaN values.")

        self.embedding = nn.Embedding.from_pretrained(
            embedding_matrix.type(torch.float32), freeze=freeze
        )
        if freeze:
            self.embedding.weight.requires_grad = False

        # Add MLP
        if mlp_hidden_dim is None:
            mlp_hidden_dim = output_dim

        self.mlp = MLP(
            self.embedding_dim,
            mlp_hidden_dim,
            output_dim,
            num_layers=2,
            dropout=dropout,
            use_layer_norm=normalize,
        )

    def forward(self, x, embed_only=False):
        embeddings = self.embedding(x.to(next(self.parameters()).device))
        if embed_only:
            return embeddings
        x = self.mlp(embeddings)
        return x


class MLP(nn.Module):
    def __init__(
        self,
        input_dim,
        hidden_dim,
        output_dim,
        num_layers=2,
        dropout=0.1,
        use_layer_norm=True,
    ):
        super().__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.output_dim = output_dim
        self.num_layers = num_layers
        self.dropout = dropout
        self.use_layer_norm = use_layer_norm

        layers = []
        for i in range(num_layers):
            if i == 0:
                layers.append(nn.Linear(input_dim, hidden_dim))
            elif i == num_layers - 1:
                layers.append(nn.Linear(hidden_dim, output_dim))
            else:
                layers.append(nn.Linear(hidden_dim, hidden_dim))

            if (
                i < num_layers - 1
            ):  # Don't add activation and normalization after the last layer
                layers.append(nn.ReLU())
                if use_layer_norm:
                    layers.append(nn.LayerNorm(hidden_dim))
                layers.append(nn.Dropout(dropout))

        self.layers = nn.Sequential(*layers)

    def forward(self, x):
        return self.layers(x)



from functools import partial
from helical.models.base_models import HelicalDNAModel
from helical.models.evo_2 import Evo2Config
from datasets import Dataset
from .evo2_tokenizer import CharLevelTokenizer

import huggingface_hub
from huggingface_hub import hf_hub_download, constants
import os

import torch
from torch.utils.data import DataLoader
import numpy as np
from tqdm import tqdm
from typing import Dict, List, Tuple, Union
from pandas import DataFrame

from .scoring import score_sequences, score_sequences_rc

import logging

LOGGER = logging.getLogger(__name__)

try:
    from vortex.model.generation import generate as vortex_generate
    from vortex.model.model import StripedHyena
    from vortex.model.utils import dotdict, load_checkpoint
except:
    message = "Vortex is required for Evo 2 model. Please follow the instructions in the README to install it."
    LOGGER.error(message)
    raise ImportError(
        message
    )


class Evo2(HelicalDNAModel):
    """Evo 2 Model.

    Example
    -------
    ```python
    from helical.models.evo_2 import Evo2, Evo2Config

    evo2_config = Evo2Config(batch_size=1)

    evo2 = Evo2(configurer=evo2_config)

    sequences = ["ACGT" * 1000]

    dataset = evo2.process_data(data)

    print(evo2.max_seq_len)

    embeddings = evo2.get_embeddings(dataset)

    generate = evo2.generate(dataset)
    # Get the last embedding of each sequence
    print(embeddings["embeddings"][0][embeddings["original_lengths"][0]-1])
    print(embeddings["embeddings"][1][embeddings["original_lengths"][1]-1])
    print(embeddings["original_lengths"])

    # Print the generated sequences
    print(generate)
    ```

    Parameters
    ----------
    configurer : Evo2Config
        The configuration object for the Evo 2 model.

    Notes
    -------
    The Evo 2 model is a model for processing DNA sequences. It can be used to generate embeddings, score sequences, and generate sequences. These models are large and require sepcific hardware to run.
    """

    default_configurer = Evo2Config()

    def __init__(self, configurer: Evo2Config = default_configurer):
        super().__init__()
        self.configurer = configurer
        self.config = configurer.config

        self.tokenizer = CharLevelTokenizer(self.config["model_map"]["vocab_size"])
        self.model = self._load_evo2_model(self.config["model_map"]["model_name"])

        self.max_seq_len = self.config["model_map"]["max_seqlen"]

        self.model.eval()
        LOGGER.info("Evo 2 initialized successfully.")

        mode = "training" if self.model.training else "eval"

        LOGGER.info(
            f"'{self.config['model_map']['model_name']}' model is in '{mode}' mode, on device '{next(self.model.parameters()).device.type}'."
        )

    def process_data(self, sequences: Union[list[str], DataFrame]) -> Dataset:
        """Process the mRNA sequences and return a Dataset object.

        Parameters
        ----------
        sequences : list[str] or DataFrame
            The DNA sequences. If a DataFrame is provided, it should have a column named 'Sequence'.

        Returns
        -------
        Dataset
            The dataset object.
        """
        LOGGER.info(f"Processing data for Evo 2.")
        sequences = self.get_valid_dna_sequence(sequences, enforce_characters=False)

        self.max_seq_len = min(len(max(sequences, key=len)), self.max_seq_len)
        LOGGER.info(f"Maximum sequence length: {self.max_seq_len}")

        prepped_sequences = {}

        prepped_sequences["input_ids"] = sequences

        dataset = Dataset.from_dict(prepped_sequences)

        LOGGER.info("Successfully processed the data for Evo 2.")
        return dataset

    def get_embeddings(
        self, dataset: Dataset, embedding_layer: str = None
    ) -> np.ndarray:
        """Get the embeddings for the mRNA sequences.

        Parameters
        ----------
        dataset : Evo2Dataset
            The dataset object.
        embedding_layer : str
            The layer to extract embeddings from. If None, the default embedding layer is used, which is the last layer.

        Returns
        -------
        np.ndarray
            The embeddings array.
        """
        LOGGER.info("Started getting embeddings:")
        dataloader = DataLoader(
            dataset,
            collate_fn=self._collate_fn,
            batch_size=self.config["batch_size"],
            shuffle=False,
        )

        embeddings = {"embeddings": [], "original_lengths": []}

        if embedding_layer is None:
            embedding_layer = self.config["model_map"]["default_embedding_layer"]

        progress_bar = tqdm(dataloader, desc="Getting embeddings")
        with torch.no_grad():
            for batch in progress_bar:
                input_ids = batch["input_ids"].to(self.config["device"])

                _, layer_embedding = self(
                    input_ids, return_embeddings=True, layer_names=[embedding_layer]
                )

                embeddings["embeddings"].append(
                    np.concatenate(layer_embedding[embedding_layer].float().cpu().numpy())
                )
                embeddings["original_lengths"].append(batch["original_lengths"])

                del batch
                del layer_embedding

        LOGGER.info(f"Finished getting embeddings.")
        embeddings["original_lengths"] = np.array(embeddings["original_lengths"])
        return embeddings

    def _collate_fn(self, batch):

        input_ids = [item["input_ids"] for item in batch]
        original_lengths = [min(len(ids), self.max_seq_len) for ids in input_ids]

        input_ids = torch.tensor(
            [
                self.tokenizer.tokenize(ids)
                + [self.tokenizer.pad_id] * (self.max_seq_len - len(ids))
                for ids in input_ids
            ],
            dtype=torch.int,
        )
        

        batch_dict = {
            "input_ids": input_ids,
            "original_lengths": original_lengths,
        }

        if "labels" in batch[0]:
            batch_dict["labels"] = torch.tensor([item["labels"] for item in batch])

        return batch_dict

    def _forward(
        self,
        input_ids: torch.Tensor,
        return_embeddings: bool = False,
        layer_names=None,
    ) -> Tuple[torch.Tensor, Dict[str, torch.Tensor]]:
        """
        Forward pass with optional embedding extraction.

        Args:
            input_ids: Input token IDs
            return_embeddings: If True, returns embeddings from specified layers
            layer_names: List of layer names to extract embeddings from. Required if
                return_embeddings=True

        Returns:
            Tuple of (logits, embeddings_dict) if return_embeddings=True
            Tuple of (logits, None) otherwise
        """
        embeddings = {}
        handles = []

        if return_embeddings:
            if layer_names is None:
                raise ValueError(
                    "layer_names must be specified when return_embeddings=True. Look at "
                    "evo2_model.model.state_dict().keys() to see available layers."
                )

            def hook_fn(layer_name):
                def hook(_, __, output):
                    if isinstance(output, tuple):
                        output = output[0]
                    embeddings[layer_name] = output.detach()

                return hook

            # Register hooks for requested layers
            for name in layer_names:
                layer = self.model.get_submodule(name)
                handles.append(layer.register_forward_hook(hook_fn(name)))

        try:
            # Original forward pass
            with torch.no_grad():
                logits = self.model.forward(input_ids)

            if return_embeddings:
                return logits, embeddings
            return logits, None

        finally:
            for handle in handles:
                handle.remove()

    def __call__(self, input_ids, return_embeddings=False, layer_names=None):
        return self._forward(input_ids, return_embeddings, layer_names)

    def score_sequences(
        self,
        seqs: List[str],
        batch_size: int = 1,
        prepend_bos: bool = False,
        reduce_method: str = "mean",
        average_reverse_complement: bool = False,
    ) -> List[float]:
        """
        Score a list of sequences.

        Parameters
        ----------
        seqs : List[str]
            List of sequences to score.
        batch_size : int
            Batch size for scoring.
        prepend_bos : bool
            Whether to prepend the BOS token.
        reduce_method : str
            Method to reduce sequence scores.
        average_reverse_complement : bool
            Whether to average the forward and reverse complement scores.

        Returns
        ----------
        List[float]
            List of scores for the input sequences.
        """
        scoring_func = partial(
            score_sequences_rc if average_reverse_complement else score_sequences,
            model=self.model,
            tokenizer=self.tokenizer,
            batch_size=batch_size,
            prepend_bos=prepend_bos,
            reduce_method=reduce_method,
        )

        with torch.no_grad():
            try:
                scores = scoring_func(seqs)
            except Exception as e:
                raise RuntimeError(f"Error during sequence scoring: {str(e)}") from e

        return scores

    def generate(
        self,
        dataset: List[str] | DataFrame,
        n_tokens: int = 500,
        temperature: float = 1.0,
        top_k: int = 4,
        top_p: float = 1.0,
        batched: bool = True,
        cached_generation: bool = True,
        verbose: int = 0,
        force_prompt_threshold: int = None,
    ) -> Tuple[List[str], List[float]]:
        """
        Generate sequences from a processed Helical Evo 2 dataset.

        Notes
        ----------
        force_prompt_threshold: If specified, avoids OOM errors through teacher forcing if the prompt is longer than this threshold.
        If force_prompt_threshold is none, sets default assuming 1xH100 (evo2_7b) and 2xH100 (evo2_40b) to help avoid OOM errors.

        Parameters
        ----------
        prompt_seqs :  Dataset
            The dataset object.
        n_tokens : int
            Number of tokens to generate.
        temperature : float
            Temperature for sampling.
        top_k : int
            Top-k sampling.
        top_p : float
            Top-p sampling.
        batched : bool
            Whether to generate sequences in batches.
        cached_generation : bool
            Whether to cache generation.
        verbose : int
            Verbosity level.
        force_prompt_threshold : int
            Threshold for teacher forcing.

        Returns
        ----------
        Tuple[List[str], List[float]]
            List of generated sequences and their scores.
        """

        with torch.no_grad():
            output = vortex_generate(
                prompt_seqs=dataset["input_ids"],
                model=self.model,
                tokenizer=self.tokenizer,
                n_tokens=n_tokens,
                temperature=temperature,
                top_k=top_k,
                top_p=top_p,
                batched=batched,
                cached_generation=cached_generation,
                verbose=verbose,
                force_prompt_threshold=force_prompt_threshold,
            )
            return output

    def _load_evo2_model(self, model_name: str = None):
        """
        Load HuggingFace checkpoint using StripedHyena 2.
        """

        hf_model_name = self.config["model_map"]["model_hf_name"]
        filename = f"{model_name}.pt"

        # First try normal download
        try:
            weights_path = hf_hub_download(
                repo_id=hf_model_name,
                filename=filename,
            )
        # If file is split, download and join parts
        except:
            print(f"Loading checkpoint shards for {filename}")
            # If file is split, get the first part's directory to use the same cache location
            weights_path = os.path.join(
                os.path.dirname(constants.HF_HUB_CACHE), filename
            )
            if os.path.exists(weights_path):
                print(f"Found {filename}")
            else:
                # Download and join parts
                parts = []
                part_num = 0
                while True:
                    try:
                        part_path = hf_hub_download(
                            repo_id=hf_model_name, filename=f"{filename}.part{part_num}"
                        )
                        parts.append(part_path)
                        part_num += 1
                    except huggingface_hub.errors.EntryNotFoundError:
                        break

                # Join in the same directory
                with open(weights_path, "wb") as outfile:
                    for part in parts:
                        with open(part, "rb") as infile:
                            while True:
                                chunk = infile.read(8192 * 1024)
                                if not chunk:
                                    break
                                outfile.write(chunk)

                # Cleaning up the parts
                for part in parts:
                    try:
                        os.remove(part)
                    except OSError as e:
                        print(f"Error removing {part}: {e}")
                    print("Cleaned up shards, final checkpoint saved to", weights_path)

        global_config = dotdict(self.config["model_map"])
        model = StripedHyena(global_config)
        load_checkpoint(model, weights_path)

        return model



import logging
import anndata
import torch
from hydra.utils import instantiate
from torch.utils.data import DataLoader
from helical.models.transcriptformer.data.dataloader import AnnDataset
from helical.models.transcriptformer.model_dir.embedding_surgery import (
    change_embedding_layer,
)
from helical.models.transcriptformer.tokenizer.vocab import load_vocabs_and_embeddings
from helical.models.transcriptformer.utils.utils import stack_dict
from helical.models.base_models import HelicalRNAModel
from helical.utils.downloader import Downloader
from omegaconf import OmegaConf
import json
import os
import pandas as pd
from helical.constants.paths import CACHE_DIR_HELICAL
from helical.models.transcriptformer.transcriptformer_config import (
    TranscriptFormerConfig,
)
from tqdm import tqdm

torch.set_float32_matmul_precision("high")

logger = logging.getLogger(__name__)


class TranscriptFormer(HelicalRNAModel):
    """
    TranscriptFormer model for RNA-seq data.
    This class is a wrapper around the TranscriptFormer model and provides methods for
    processing data, getting embeddings, and loading the model.

    Parameters
    ----------
    configurer: TranscriptFormerConfig
        Configuration object for the TranscriptFormer model.

    Example
    -------
    ```python
    from helical.models.transcriptformer.model import TranscriptFormer
    from helical.models.transcriptformer.transcriptformer_config import TranscriptFormerConfig
    import anndata as ad

    configurer = TranscriptFormerConfig()
    model = TranscriptFormer(configurer)

    ann_data = ad.read_h5ad("/path/to/data.h5ad")

    dataset = model.process_data([ann_data])
    embeddings = model.get_embeddings(dataset)
    print(embeddings)
    ```

    Notes
    -----
    We use the implementation from [this repository](https://github.com/czi-ai/transcriptformer),
    which comes from the original authors. You can find the description of the method in
    [this paper](https://www.biorxiv.org/content/10.1101/2025.04.25.650731v1).
    """

    configurer = TranscriptFormerConfig()

    def __init__(self, configurer: TranscriptFormerConfig = configurer):
        super().__init__(configurer.config)
        self.config = configurer.config

        downloader = Downloader()
        for file in configurer.list_of_files_to_download:
            downloader.download_via_name(file)

        logger.info(f"Loading cache config for {configurer.model_name}")
        cache_config_path = os.path.join(
            CACHE_DIR_HELICAL, "transcriptformer", configurer.model_name, "config.json"
        )
        with open(cache_config_path) as f:
            cache_config_dict = json.load(f)
        cache_config = OmegaConf.create(cache_config_dict)

        # Merge the cache config with the config provided by the user
        self.config = OmegaConf.merge(cache_config, self.config)

        self.config.model.inference_config.load_checkpoint = os.path.join(
            CACHE_DIR_HELICAL,
            "transcriptformer",
            configurer.model_name,
            "model_weights.pt",
        )
        self.config.model.data_config.aux_vocab_path = os.path.join(
            CACHE_DIR_HELICAL, "transcriptformer", configurer.model_name, "vocabs"
        )
        self.config.model.data_config.aux_cols = "assay"
        self.config.model.data_config.esm2_mappings_path = os.path.join(
            CACHE_DIR_HELICAL, "transcriptformer", configurer.model_name, "vocabs"
        )
        logger.info(f"Merged cache config with user config for {configurer.model_name}")

        logger.debug(OmegaConf.to_yaml(self.config))

        # Load vocabs and embeddings
        (self.gene_vocab, self.aux_vocab), self.emb_matrix = load_vocabs_and_embeddings(
            self.config
        )

        # Instantiate the model
        logger.info("Instantiating the model")
        self.model = instantiate(
            self.config.model,
            gene_vocab_dict=self.gene_vocab,
            aux_vocab_dict=self.aux_vocab,
            emb_matrix=self.emb_matrix,
            emb_mode=self.config.model.inference_config.emb_mode,
        )
        self.model.eval()

        logger.info("Model instantiated successfully")

        # Check if checkpoint is supplied
        if (
            not hasattr(self.model.inference_config, "load_checkpoint")
            or not self.model.inference_config.load_checkpoint
        ):
            raise ValueError(
                "No checkpoint provided for inference. Please specify a checkpoint path in "
                "model.inference_config.load_checkpoint"
            )

        logger.info("Loading model checkpoint")
        state_dict = torch.load(
            self.model.inference_config.load_checkpoint, weights_only=True
        )

        # Filter out auxiliary embedding weights if aux_vocab_path is None
        if self.model.data_config.aux_vocab_path is None:
            filtered_state_dict = {
                k: v
                for k, v in state_dict.items()
                if not k.startswith("aux_embeddings.")
            }
            state_dict = filtered_state_dict

        self.model.load_state_dict(state_dict)
        logger.info("Model weights loaded successfully")

        # Perform embedding surgery if specified in config
        if self.model.inference_config.pretrained_embedding is not None:
            logger.info("Performing embedding surgery")
            # Check if pretrained_embedding_paths is a list, if not convert it to a list
            if not isinstance(self.model.inference_config.pretrained_embedding, list):
                pretrained_embedding_paths = [
                    self.model.inference_config.pretrained_embedding
                ]
            else:
                pretrained_embedding_paths = (
                    self.model.inference_config.pretrained_embedding
                )
            self.model, self.gene_vocab = change_embedding_layer(
                self.model, pretrained_embedding_paths
            )

        mode = "training" if self.model.training else "eval"
        logger.info(
            f"TranscriptFormer '{configurer.model_name}' model is in '{mode}' mode, "
            f"on device GPU generating embeddings for '{self.model.inference_config.output_keys}' & {self.model.emb_mode} embeddings."
        )

    def process_data(self, data_files: list[str] | list[anndata.AnnData]):
        """
        Process the data for TranscriptFormer.

        Parameters
        ----------
        data_files: list[str] | list[anndata.AnnData]
            List of paths to AnnData files or AnnData objects.

        Returns
        -------
            dataset: The processed dataset.
        """
        # Load dataset

        logger.info(f"Processing data for TranscriptFormer.")

        data_kwargs = {
            "gene_vocab": self.gene_vocab,
            "aux_vocab": self.aux_vocab,
            "max_len": self.model.model_config.seq_len,
            "pad_zeros": self.model.data_config.pad_zeros,
            "pad_token": self.model.data_config.gene_pad_token,
            "sort_genes": self.model.data_config.sort_genes,
            "filter_to_vocab": self.model.data_config.filter_to_vocabs,
            "filter_outliers": self.model.data_config.filter_outliers,
            "gene_col_name": self.model.data_config.gene_col_name,
            "normalize_to_scale": self.model.data_config.normalize_to_scale,
            "randomize_order": self.model.data_config.randomize_genes,
            "min_expressed_genes": self.model.data_config.min_expressed_genes,
            "clip_counts": self.model.data_config.clip_counts,
            "obs_keys": self.model.inference_config.obs_keys,
        }
        dataset = AnnDataset(data_files, **data_kwargs)
        return dataset

    def get_embeddings(self, dataset: AnnDataset) -> torch.Tensor:
        """
        Get the embeddings for the dataset.

        Parameters
        ----------
        dataset: AnnDataset
            The dataset to get the embeddings for.

        Returns
        -------
        embeddings: torch.Tensor
            The embeddings for the dataset.
        """
        # Create dataloader
        dataloader = DataLoader(
            dataset,
            batch_size=self.model.inference_config.batch_size,
            num_workers=self.model.data_config.n_data_workers,
            drop_last=False,
            pin_memory=True,
            collate_fn=dataset.collate_fn,
        )

        output = []
        progress_bar = tqdm(
            dataloader,
            desc="Embedding Cells",
            total=len(dataloader),
            unit="batch",
        )

        self.model.to(self.config.model.inference_config.device)
        with torch.no_grad():
            for batch in progress_bar:
                output_batch = self.model.inference(batch)
                output.append(output_batch)

        logger.info("Combining predictions")
        concat_output = stack_dict(output)

        # Create pandas DataFrames from the obs and uns data in concat_output
        obs_df = pd.DataFrame(concat_output["obs"])
        uns = (
            {"llh": pd.DataFrame({"llh": concat_output["llh"]})}
            if "llh" in concat_output
            else None
        )
        obsm = {}

        # Add all other output keys to the obsm
        for k in self.model.inference_config.output_keys:
            if k in concat_output:
                obsm[k] = concat_output[k].numpy()

        # Create a new AnnData object with the embeddings
        self.output_adata = anndata.AnnData(
            obs=obs_df,
            obsm=obsm,
            uns=uns,
        )

        logger.info(f"Returning '{self.model.emb_mode}_embeddings' from output.")
        embeddings = concat_output[self.model.emb_mode + "_embeddings"]
        return embeddings

    def get_output_adata(self) -> anndata.AnnData:
        """
        Get the output AnnData object. Only call this after running 'get_embeddings'.

        Returns
        -------
            output_adata: The output AnnData object.
        """
        if not hasattr(self, "output_adata"):
            message = (
                "Output AnnData object not found. Please run 'get_embeddings' first."
            )
            logger.error(message)
            raise ValueError(message)
        logger.info(
            "Returning output AnnData object, embeddings are stored in .obsm['embeddings'], uns['llh'] contains the log-likelihoods"
        )
        return self.output_adata



import numpy as np
from helical.models.scgpt.binning import _digitize, binning
import torch
import pytest


def test_digitize_basic():
    np.random.seed(42)
    x = np.array([-99, 1, 2, 3, 4, 5, 6, 9])
    bins = np.array([2, 4, 9])
    result = _digitize(x, bins)
    # -99 and 1 are below the first bin, 2 and 3 are in the first bin
    # 4, 5 and 6 are in the second bin, 9 in the third
    expected = np.array([0, 0, 1, 1, 2, 2, 2, 3])
    assert np.array_equal(result, expected)


def test_digitize_side_one():
    np.random.seed(42)
    x = np.array([1, 2, 3, 4, 5])
    bins = np.array([2, 4])
    result = _digitize(x, bins, side="one")
    expected = np.array([0, 1, 1, 2, 2])
    assert np.array_equal(result, expected)


def test_digitize_empty_array():
    np.random.seed(42)
    x = np.array([])
    bins = np.array([2, 4])
    result = _digitize(x, bins)
    expected = np.array([])
    assert np.array_equal(result, expected)


def test_digitize_identical_bins():
    np.random.seed(42)
    x = np.array([1, 2, 3, 4, 5])
    bins = np.array([2, 2, 4, 4])
    result = _digitize(x, bins)
    expected = np.array([0, 1, 2, 3, 4])
    assert np.array_equal(result, expected)


@pytest.mark.parametrize(
    "row, expected",
    (
        (np.array([1, 2, 3, 4, 5, 6, 7, 8, 9]), np.array([1, 1, 1, 2, 2, 2, 3, 3, 4])),
        (
            torch.tensor([1, 2, 3, 4, 5, 6, 7, 8, 9]),
            np.array([1, 1, 1, 2, 2, 2, 3, 3, 4]),
        ),
        # distrubution of the bins depends on the distribution of the data
        (np.array([1, 1, 1, 1, 1, 1, 1, 8, 9]), np.array([2, 1, 1, 2, 3, 3, 3, 3, 4])),
        (np.array([1, 2, 1, 1, 9, 6, 7, 8, 9]), np.array([1, 2, 1, 1, 4, 2, 2, 3, 4])),
    ),
)
def test_binning_basic(row, expected):
    np.random.seed(42)
    n_bins = 5
    result = binning(row, n_bins)
    assert np.array_equal(result, expected)


def test_binning_with_zeros():
    np.random.seed(42)
    row = np.array([0, 0, 0, 0, 0])
    n_bins = 5
    result = binning(row, n_bins)
    expected = np.zeros_like(row)
    assert np.array_equal(result, expected)


def test_binning_with_negative_values():
    np.random.seed(42)
    row = np.array([-1, -2, -3, -4, -5, -6, -7, -8, -9])
    n_bins = 5
    result = binning(row, n_bins)
    expected = np.array([4, 3, 3, 2, 2, 2, 1, 1, 1])
    assert np.array_equal(result, expected)



import pandas as pd
import anndata as ad
import numpy as np
from scipy.sparse import lil_matrix
from datasets import Dataset
import logging

LOGGER = logging.getLogger(__name__)


def get_anndata_from_hf_dataset(dataset: Dataset) -> ad.AnnData:
    """
    Convert a HuggingFace dataset to an AnnData object
    It assumes that the dataset has the following columns:
    - raw_counts (specifying the raw counts of the data, excluding 0s)
    - rows (specifying the row indices of the raw_counts)
    - Any other column that will be used as obs

    The id of the raw_counts are the gene names as a string separated by commas.

    Parameters
    ----------
    dataset : Dataset
        A HuggingFace dataset object

    Returns
    -------
    AnnData
        An AnnData object containing the data from the input dataset
    """
    # obs
    excluded_features = ["raw_counts", "rows", "size"]
    observation_names = [
        obs for obs in dataset.features.keys() if obs not in excluded_features
    ]
    obs_data = pd.DataFrame(
        dataset.select_columns(observation_names).data.to_pandas(),
        columns=observation_names,
    )

    # raw counts
    lil = lil_matrix((len(dataset), dataset[0]["size"]))
    lil.data = np.array(dataset["raw_counts"], dtype="object")
    lil.rows = np.array(dataset["rows"], dtype="object")

    # create AnnData object
    ann_data = ad.AnnData(lil.tocsr(), obs=obs_data)

    # add gene names
    var_names = dataset.features["raw_counts"].id.split(",")
    if len(var_names) != ann_data.shape[1]:
        message = f"Number of gene names ({len(var_names)}) does not match the number of genes ({ann_data.shape[1]})"
        LOGGER.error(message)
        raise ValueError(message)

    ann_data.var_names = [name.upper() for name in var_names]
    ann_data.var["gene_name"] = ann_data.var_names

    return ann_data



import pytest

import csv
from typing import List, Union
import numpy as np
import torch
import torch.nn.functional as F

try:
    from helical.models.evo_2 import Evo2, Evo2Config

    evo2_unavailable = False  # only run tests if able to import the package
except:
    evo2_unavailable = True

torch.manual_seed(1)
torch.cuda.manual_seed(1)


@pytest.mark.skipif(evo2_unavailable, reason="No Evo 2 module present")
@pytest.fixture
def read_prompts() -> Union[List[List[str]]]:
    """Read prompts from input file."""
    promptseqs: List[str] = []

    with open("ci/tests/data/prompts.csv", encoding="utf-8-sig", newline="") as csvfile:
        reader = csv.reader(csvfile)
        next(reader)  # Skip header
        for row in reader:
            promptseqs.append(row[0])

    return promptseqs


@pytest.mark.skipif(evo2_unavailable, reason="No Evo 2 module present")
@pytest.fixture
def evo2_model():
    """Initialize Evo2 model."""
    model_config = Evo2Config(model_name="evo2-7b", batch_size=1)
    model = Evo2(model_config)
    return model


@pytest.mark.skipif(evo2_unavailable, reason="No Evo 2 module present")
@pytest.fixture
def test_forward_pass(evo2_model, read_prompts):
    """Test model forward pass accuracy on sequences."""
    sequences = read_prompts
    losses = []
    accuracies = []

    for seq in sequences:
        # Convert sequence to model input format
        input_ids = torch.tensor(evo2_model.tokenizer.tokenize(seq), dtype=int).to(
            "cuda:0"
        )

        with torch.inference_mode():
            # Forward pass
            logits, _ = evo2_model.model.forward(input_ids.unsqueeze(0))

            # Calculate loss and accuracy
            target_ids = input_ids[1:]  # Shift right for next token prediction
            pred_logits = logits[0, :-1, :]

            # Cross entropy loss
            loss = F.cross_entropy(pred_logits, target_ids.long())

            # Get predictions
            pred_tokens = torch.argmax(pred_logits, dim=-1)

            # Calculate accuracy
            accuracy = (target_ids == pred_tokens).float().mean().item()

            losses.append(loss.item())
            accuracies.append(accuracy)

    # Print sequence results
    print("\nSequence Results:")
    for i, (loss, acc) in enumerate(zip(losses, accuracies)):
        print(f"Sequence {i+1}: Loss = {loss:.3f}, Accuracy = {acc:.2%}")
        if acc < 0.5:
            print(
                "WARNING: Forward pass accuracy is below 50% on test sequence. Model may be broken, trained models should have >80% accuracy."
            )

    return accuracies, losses


@pytest.mark.skipif(evo2_unavailable, reason="No Evo 2 module present")
def test_evo2(test_forward_pass, evo2_model):
    """
    Test sequence prediction accuracy using Evo2 models.
    Expected results for forward pass:
    - Evo 2 40B 1m: Loss ~0.216, Accuracy ~91.67%
    - Evo 2 7B 1m: Loss ~0.348, Accuracy ~86.35%
    - Evo 2 1B base: Loss ~0.502, Accuracy ~79.56%
    """
    accuracies, losses = test_forward_pass
    # Set random seeds

    # Calculate and validate results
    mean_loss = np.mean(losses)
    mean_accuracy = np.mean(accuracies) * 100
    print(f"\nMean Loss: {mean_loss:.3f}")
    print(f"Mean Accuracy: {mean_accuracy:.3f}%")

    # Validate against expected scores
    eps = 1e-3  # epsilon for float comparison
    expected_metrics = {
        # "evo2_40b": {"loss": 0.2159424, "acc": 91.673},
        "evo2_7b": {"loss": 0.3476563, "acc": 86.346},
        # "evo2_40b_base": {"loss": 0.2149658, "acc": 91.741},
        # "evo2_7b_base": {"loss": 0.3520508, "acc": 85.921},
        # "evo2_1b_base": {"loss": 0.501953125, "acc": 79.556},
    }

    expected = expected_metrics[evo2_model.config["model_map"]["model_name"]]
    if abs(mean_loss - expected["loss"]) < eps:
        print(f"\nTest Passed! Loss matches expected {expected['loss']:.3f}")
    else:
        print(
            f"\nTest Failed: Expected loss {expected['loss']:.3f}, got {mean_loss:.3f}"
        )



import requests
from helical.utils.logger import Logger
from helical.constants.enums import LoggingType, LoggingLevel
import logging
import sys
from pathlib import Path
from tqdm import tqdm
from helical.constants.paths import CACHE_DIR_HELICAL
import hashlib

LOGGER = logging.getLogger(__name__)
INTERVAL = 1000
CHUNK_SIZE = 1024 * 1024 * 10
LOADING_BAR_LENGTH = 50


class Downloader(Logger):
    def __init__(
        self, loging_type=LoggingType.CONSOLE, level=LoggingLevel.INFO
    ) -> None:
        super().__init__(loging_type, level)
        self.display = True
        self.session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(pool_maxsize=100, pool_connections=100)
        self.session.mount("https://", adapter)

    def download_via_link(self, output: Path, link: str) -> None:
        if output.is_file():
            LOGGER.debug(
                f"File: '{output}' exists already. File is not overwritten and nothing is downloaded."
            )
            return

        LOGGER.info(f"Starting to download: '{link}'")
        response = self.session.get(link, stream=True)

        if response.status_code != 200:
            message = f"Failed downloading file from '{link}' with status code: {response.status_code}"
            LOGGER.error(message)
            raise Exception(message)

        total_length = response.headers.get("content-length")
        self.data_length = 0
        self.total_length = int(total_length) if total_length else None

        output.parent.mkdir(parents=True, exist_ok=True)

        if not self.total_length:
            # If total length unknown, just write content directly without progress bar
            with open(output, "wb") as f:
                f.write(response.content)
        else:
            with (
                open(output, "wb") as f,
                tqdm(
                    total=self.total_length, unit="B", unit_scale=True, desc=output.name
                ) as pbar,
            ):
                for data in response.iter_content(chunk_size=CHUNK_SIZE):
                    f.write(data)
                    pbar.update(len(data))

    def download_via_name(self, name: str) -> None:
        bucket_name = "helicalpackage"
        region = "eu-west-2"
        s3_key = name
        output = Path(CACHE_DIR_HELICAL) / name
        url = f"https://{bucket_name}.s3.{region}.amazonaws.com/{s3_key}"

        output.parent.mkdir(parents=True, exist_ok=True)

        if not self.check_file_valid(url, output):
            if output.is_file():
                LOGGER.warning(
                    f"File '{output}' is corrupted or invalid. Deleting and re-downloading."
                )
                output.unlink()  # delete corrupted file

            LOGGER.info(f"Downloading '{name}'")
            self.download_via_link(output, url)

            # Validate again after download
            if not self.check_file_valid(url, output):
                raise ValueError(f"File '{output}' failed validation after download.")
            else:
                LOGGER.info(f"File saved to: '{output}'")
        else:
            LOGGER.debug(f"File '{output}' already exists and is valid.")

    def check_file_valid(self, url: str, file_path: Path) -> bool:
        """
        Validates a local file against the remote object's ETag or Content-Length.
        Returns True if file exists and is considered valid.
        """
        try:
            if not file_path.is_file():
                return False

            head = self.session.head(url)
            etag = head.headers.get("ETag", "").strip('"')
            remote_size = int(head.headers.get("Content-Length", 0))
            local_size = file_path.stat().st_size

            if local_size != remote_size:
                LOGGER.warning(
                    f"File size mismatch: local {local_size}, remote {remote_size}"
                )
                return False

            # Handle single-part ETag (true MD5)
            if "-" not in etag:
                local_md5 = self._md5_checksum(file_path)
                if etag != local_md5:
                    LOGGER.warning(
                        f"MD5 checksum mismatch: local {local_md5}, remote {etag}"
                    )
                    return False

            # For multi-part uploads (ETag has "-"), assume size check is sufficient
            return True

        except Exception as e:
            LOGGER.warning(f"Validation failed for {file_path}: {e}")
            return False

    def _md5_checksum(self, filename: Path, chunk_size: int = 1024 * 1024) -> str:
        m = hashlib.md5()
        with open(filename, "rb") as f:
            for chunk in iter(lambda: f.read(chunk_size), b""):
                m.update(chunk)
        return m.hexdigest()



from helical.models.helix_mrna import HelixmRNAFineTuningModel, HelixmRNAConfig
import hydra
from omegaconf import DictConfig


@hydra.main(
    version_base=None,
    config_path="../run_models/configs",
    config_name="helix_mrna_config",
)
def run_fine_tuning(cfg: DictConfig):
    input_sequences = ["EACU" * 20, "EAUG" * 20, "EAUG" * 20, "EACU" * 20, "EAUU" * 20]
    labels = [0, 2, 2, 0, 1]

    helix_mrna_config = HelixmRNAConfig(**cfg)
    helix_mrna_fine_tune = HelixmRNAFineTuningModel(
        helix_mrna_config=helix_mrna_config,
        fine_tuning_head="classification",
        output_size=3,
    )

    train_dataset = helix_mrna_fine_tune.process_data(input_sequences)

    helix_mrna_fine_tune.train(train_dataset=train_dataset, train_labels=labels)

    outputs = helix_mrna_fine_tune.get_outputs(train_dataset)
    print(outputs)

    # save and load model
    helix_mrna_fine_tune.save_model("./helix_mrna_fine_tuned_model.pt")
    helix_mrna_fine_tune.load_model("./helix_mrna_fine_tuned_model.pt")

    outputs = helix_mrna_fine_tune.get_outputs(train_dataset)
    print(outputs)

if __name__ == "__main__":
    run_fine_tuning()



import torch

from helical.models.transcriptformer.model_dir.model import Transcriptformer
from helical.models.transcriptformer.tokenizer.vocab import (
    SPECIAL_TOKENS,
    construct_gene_embeddings,
)


def change_embedding_layer(
    model: Transcriptformer,
    pretrained_embedding_paths: list[str],
    special_tokens: list | None = None,
):
    """
    Change the embedding layer of a model to a new embedding layer.

    Args:
        model (torch.nn.Module): The model to change the embedding layer of.
        pretrained_embedding_paths (list[str]): The paths to the pretrained embedding files.
        special_tokens (Optional[list]): The special tokens to add to the model.


    Returns
    -------
        torch.nn.Module: The model with the new embedding layer.
    """
    if special_tokens is None:
        special_tokens = SPECIAL_TOKENS

    old_special_tokens = [i for i in special_tokens if i in model.gene_vocab.vocab_dict]

    # Extract the rows of the embedding matrix that correspond to the special tokens
    old_special_token_indices = torch.tensor(
        [model.gene_vocab.vocab_dict[token] for token in old_special_tokens],
        dtype=torch.int64,
    )
    # Move model to GPU - needed if model is not re-loaded from checkpoint
    model = model.to("cuda")
    old_special_token_embeddings = model.gene_embeddings.embedding(
        old_special_token_indices.to("cuda")
    )

    # Read the new embeddings from the files
    gene_vocab, new_embedding_matrix = construct_gene_embeddings(
        pretrained_embedding_paths,
        [token for token in special_tokens if token not in old_special_tokens],
    )

    # Concatenate the old special token embeddings with the new embeddings
    new_embedding_matrix = torch.cat(
        [old_special_token_embeddings, torch.Tensor(new_embedding_matrix).to("cuda")],
        dim=0,
    )

    # Update the vocab indices of the model
    gene_vocab = {
        gene: idx + len(old_special_tokens) for gene, idx in gene_vocab.items()
    }

    # Update the gene_vocab with the special tokens
    gene_vocab.update({token: idx for idx, token in enumerate(old_special_tokens)})

    # Create a new embedding layer
    new_embedding = torch.nn.Embedding(
        len(gene_vocab),
        new_embedding_matrix.shape[1],
        padding_idx=model.gene_vocab.pad_idx,
    )

    # Set device to be the same as the model's embedding layer
    new_embedding = new_embedding.to(model.gene_embeddings.embedding.weight.device)

    # Assign the weights to the new embedding layer, from gene_vocab
    new_embedding.weight.data.copy_(new_embedding_matrix)

    # Replace the old embedding layer with the new one
    model.gene_embeddings.embedding = new_embedding

    # Update the gene_vocab
    model.gene_vocab.vocab_dict = gene_vocab
    model.gene_vocab.embedding_matrix = new_embedding_matrix

    return model, gene_vocab



from typing import Optional, Literal
from helical.constants.paths import CACHE_DIR_HELICAL
from pathlib import Path


class scGPTConfig:
    """
    Configuration class to use the scGPT Model.

    Parameters
    ----------
    pad_token : str, optional, default="<pad>"
        The padding token
    batch_size : int, optional, default=24
        The batch size
    emb_mode : Literal["cls", "cell", "gene"], optional, default="cell"
        The embedding mode to use. "cls" uses the cls token, "cell" uses the cell token, "gene" uses the gene token.
    nlayers : int, optional, default=12
        The number of layers
    nheads : int, optional, default=8
        The number of heads
    embsize : int, optional, default=512
        The embedding size
    d_hid : int, optional, default=512
        The hidden dimension
    dropout : float, optional, default=0.2
        The dropout rate
    n_layers_cls : int, optional, default=3
        The number of classification layers
    mask_value : int, optional, default=-1
        The mask value
    pad_value : int, optional, default=-2
        The padding value
    world_size : int, optional, default=8
        The world size
    accelerator : bool, optional, default=False
        The accelerator configuration. By default same device as model.
    device : Literal["cpu", "cuda"], optional, default="cpu"
        The device to use. Either use "cuda" or "cpu".
    binning_seed : int, optional, default=123
        The seed for the binning

    Returns
    -------
    scGPTConfig
        The scGPT configuration object

    Notes
    -----
    This configuration contains all the default parameteres that have been used in the original scGPT repository.

    """

    def __init__(
        self,
        pad_token: str = "<pad>",
        batch_size: int = 24,
        emb_mode: Literal["cls", "cell", "gene"] = "cell",
        nlayers: int = 12,
        nheads: int = 8,
        embsize: int = 512,
        d_hid: int = 512,
        dropout: float = 0.2,
        n_layers_cls: int = 3,
        mask_value: int = -1,
        pad_value: int = -2,
        world_size: int = 8,
        accelerator: Optional[bool] = False,
        device: Literal["cpu", "cuda"] = "cpu",
        binning_seed: int = 123,
    ):

        model_name = "best_model"  # TODO: Include more models
        list_of_files_to_download = [
            "scgpt/scGPT_CP/vocab.json",
            f"scgpt/scGPT_CP/{model_name}.pt",
        ]
        model_path = Path(CACHE_DIR_HELICAL, "scgpt/scGPT_CP", f"{model_name}.pt")

        self.config = {
            "model_path": model_path,
            "list_of_files_to_download": list_of_files_to_download,
            "pad_token": pad_token,
            "batch_size": batch_size,
            "emb_mode": emb_mode,
            "nlayers": nlayers,
            "nheads": nheads,
            "embsize": embsize,
            "d_hid": d_hid,
            "dropout": dropout,
            "n_layers_cls": n_layers_cls,
            "mask_value": mask_value,
            "pad_value": pad_value,
            "world_size": world_size,
            "accelerator": accelerator,
            "device": device,
            "MAX_LENGTH": 1200,
            "binning_seed": binning_seed,
        }



from typing import Literal
from pathlib import Path
from helical.constants.paths import CACHE_DIR_HELICAL


class HyenaDNAConfig:
    """
    Configuration class for Hyena DNA model.

    Parameters
    ----------
    model_name : Literal["hyenadna-tiny-1k-seqlen", "hyenadna-tiny-1k-seqlen-d256"], optional, default="hyenadna-tiny-1k-seqlen"
        The name of the model.
    batch_size : int, optional, default=5
        The batch size to use for all tasks.
    n_layer : int, optional, default=2
        The number of layers in the model.
    vocab_size : int, optional, default=12
        The size of the vocabulary.
    resid_dropout : float, optional, default=0.0
        The dropout rate for residual connections.
    embed_dropout : float, optional, default=0.1
        The dropout rate for embedding layer.
    fused_mlp : bool, optional, default=False
        Whether to use fused MLP.
    fused_dropout_add_ln : bool, optional, default=True
        Whether to use fused dropout and layer normalization.
    residual_in_fp32 : bool, optional, default=True
        Whether to use FP32 for residual connections.
    checkpoint_mixer : bool, optional, default=False
        Whether to use checkpointing for mixer layers.
    checkpoint_mlp : bool, optional, default=False
        Whether to use checkpointing for MLP layers.
    pad_vocab_size_multiple : int, optional, default=8
        The multiple to pad the vocabulary size.
    return_hidden_state : bool, optional, default=True
        Whether to return the hidden state.
    device : Literal["cpu", "cuda"], optional, default="cpu"
        The device to use.
    layer : dict, optional, default={"_name_": "hyena", "emb_dim": 5, "filter_order": 64, "local_order": 3, "l_max": 1026, "modulate": True, "w": 10, "lr": 6e-4, "wd": 0.0, "lr_pos_emb": 0.0}
        Dictionary containing layer-specific parameters.

    Attributes
    ----------
    model_map : dict
        A dictionary mapping model names to their corresponding configuration parameters.
    config : dict
        A dictionary containing the configuration parameters for the Hyena DNA model.

    Raises
    ------
    ValueError
        If the specified model name is not found in the available models.

    """

    def __init__(
        self,
        model_name: Literal[
            "hyenadna-tiny-1k-seqlen", "hyenadna-tiny-1k-seqlen-d256"
        ] = "hyenadna-tiny-1k-seqlen",
        batch_size: int = 5,
        n_layer: int = 2,
        vocab_size: int = 12,
        resid_dropout: float = 0.0,
        embed_dropout: float = 0.1,
        fused_mlp: bool = False,
        fused_dropout_add_ln: bool = True,
        residual_in_fp32: bool = True,
        checkpoint_mixer: bool = False,
        checkpoint_mlp: bool = False,
        pad_vocab_size_multiple: int = 8,
        return_hidden_state: bool = True,
        device: Literal["cpu", "cuda"] = "cpu",
        layer: dict = {
            "_name_": "hyena",
            "emb_dim": 5,
            "filter_order": 64,
            "local_order": 3,
            "l_max": 1026,
            "modulate": True,
            "w": 10,
            "lr": 6e-4,
            "wd": 0.0,
            "lr_pos_emb": 0.0,
        },
    ):

        # model specific parameters
        self.model_map = {
            "hyenadna-tiny-1k-seqlen": {
                "d_model": 128,
                "d_inner": 512,
                "max_length": 1024,  # for max_length see https://github.com/HazyResearch/hyena-dna/blob/main/huggingface.py
            },
            "hyenadna-tiny-1k-seqlen-d256": {
                "d_model": 256,
                "d_inner": 1024,
                "max_length": 1024,
            },
        }

        if model_name not in self.model_map:
            raise ValueError(
                f"Model name {model_name} not found in available models: {self.model_map.keys()}"
            )

        list_of_files_to_download = [f"hyena_dna/{model_name}.ckpt"]

        self.config = {
            "model_name": model_name,
            "model_path": Path(CACHE_DIR_HELICAL, f"hyena_dna/{model_name}.ckpt"),
            "list_of_files_to_download": list_of_files_to_download,
            "batch_size": batch_size,
            "d_model": self.model_map[model_name]["d_model"],
            "n_layer": n_layer,
            "d_inner": self.model_map[model_name]["d_inner"],
            "vocab_size": vocab_size,
            "resid_dropout": resid_dropout,
            "embed_dropout": embed_dropout,
            "fused_mlp": fused_mlp,
            "fused_dropout_add_ln": fused_dropout_add_ln,
            "residual_in_fp32": residual_in_fp32,
            "checkpoint_mixer": checkpoint_mixer,
            "checkpoint_mlp": checkpoint_mlp,
            "pad_vocab_size_multiple": pad_vocab_size_multiple,
            "return_hidden_state": return_hidden_state,
            "device": device,
            "layer": layer,
            "max_length": self.model_map[model_name]["max_length"],
        }



from helical.models.hyena_dna import HyenaDNA, HyenaDNAConfig
import hydra
from omegaconf import DictConfig


@hydra.main(version_base=None, config_path="configs", config_name="hyena_dna_config")
def run(cfg: DictConfig):
    hyena_config = HyenaDNAConfig(**cfg)
    model = HyenaDNA(configurer=hyena_config)

    sequence = ["A", "CC", "TTTT", "ACGTN", "ACGT"]

    dataset = model.process_data(sequence)
    embeddings = model.get_embeddings(dataset)
    print(embeddings.shape)


if __name__ == "__main__":
    run()



import logging

try:
    from .evo_2_config import Evo2Config
    from .model import Evo2
except:
    LOGGER = logging.getLogger(__name__)
    LOGGER.info(
        "Evo 2 not available: If you want to use this model, please follow the installation instructions in the evo_2/README."
    )



from helical.models.geneformer import GeneformerConfig, GeneformerFineTuningModel
from helical.utils import get_anndata_from_hf_dataset
from datasets import load_dataset
import anndata as ad
import hydra
from omegaconf import DictConfig


@hydra.main(
    version_base=None,
    config_path="../run_models/configs",
    config_name="geneformer_config",
)
def run_fine_tuning(cfg: DictConfig):

    # either load via huggingface
    # hf_dataset = load_dataset("helical-ai/yolksac_human",split="train[:5%]", trust_remote_code=True, download_mode="reuse_cache_if_exists")
    # ann_data = get_anndata_from_hf_dataset(hf_dataset)

    # or load directly from anndata file
    ann_data = ad.read_h5ad("./yolksac_human.h5ad")

    cell_types = list(ann_data.obs["LVL1"][:10])
    label_set = set(cell_types)

    geneformer_config = GeneformerConfig(**cfg)
    geneformer_fine_tune = GeneformerFineTuningModel(
        geneformer_config=geneformer_config,
        fine_tuning_head="classification",
        output_size=len(label_set),
    )

    dataset = geneformer_fine_tune.process_data(ann_data[:10])

    dataset = dataset.add_column("cell_types", cell_types)
    class_id_dict = dict(zip(label_set, [i for i in range(len(label_set))]))

    def classes_to_ids(example):
        example["cell_types"] = class_id_dict[example["cell_types"]]
        return example

    dataset = dataset.map(classes_to_ids, num_proc=1)

    geneformer_fine_tune.train(train_dataset=dataset, label="cell_types")
    outputs = geneformer_fine_tune.get_outputs(dataset)
    print(outputs)

    # save and load model
    geneformer_fine_tune.save_model("./geneformer_fine_tuned_model.pt")
    geneformer_fine_tune.load_model("./geneformer_fine_tuned_model.pt")

    outputs = geneformer_fine_tune.get_outputs(dataset)
    print(outputs)

if __name__ == "__main__":
    run_fine_tuning()



import pytest
import numpy as np
import torch

try:
    from helical.models.caduceus import Caduceus, CaduceusConfig

    caduceus_unavailable = False  # only run tests if able to import the package
except:
    caduceus_unavailable = True


@pytest.mark.skipif(caduceus_unavailable, reason="No Caduceus module present")
@pytest.mark.parametrize(
    "model_name",
    [
        "caduceus-ph-4L-seqlen-1k-d118",
        "caduceus-ph-4L-seqlen-1k-d256",
        "caduceus-ph-16L-seqlen-131k-d256",
        "caduceus-ps-4L-seqlen-1k-d118",
        "caduceus-ps-4L-seqlen-1k-d256",
        "caduceus-ps-16L-seqlen-131k-d256",
    ],
)
def test_caduceus_valid_model_names(model_name):
    """
    Test case for the Caduceus class initialization.
    """
    CaduceusConfig(model_name=model_name)


@pytest.mark.skipif(caduceus_unavailable, reason="No Caduceus module present")
@pytest.mark.parametrize(
    "model_name",
    [
        "caduceus-pq-4L-seqlen-1k-d118",
        "caduceus-ph-9L-seqlen-1k-d256",
    ],
)
def test_caduceus_invalid_model_names(model_name):
    """
    Test case when an invalid model name is provided.
    Verifies that a ValueError is raised when an invalid model name is passed to the CaduceusConfig constructor.
    """
    with pytest.raises(ValueError):
        CaduceusConfig(model_name=model_name)


@pytest.mark.skipif(caduceus_unavailable, reason="No Caduceus module present")
def test_runtime_error_when_cuda_unavailable(mocker):
    """
    Test that an error is raised when CUDA is not available since the model can't run without it.
    """
    mocker.patch.object(torch.cuda, "is_available", return_value=False)
    with pytest.raises(RuntimeError):
        Caduceus(CaduceusConfig())


@pytest.mark.skipif(caduceus_unavailable, reason="No Caduceus module present")
@pytest.mark.parametrize(
    "input_sequence, expected_sequence",
    [
        # Valid DNA sequences
        ("A", [7, 1]),
        ("CC", [8, 8, 1]),
        ("TTTT", [10, 10, 10, 10, 1]),
        ("ACGTN", [7, 8, 9, 10, 11, 1]),
        ("ACG" * 256, [7, 8, 9] * 256 + [1]),
    ],
)
def test_caduceus_process_data(input_sequence, expected_sequence, mocker):
    """
    Test the process_data method of the Caduceus model.
    The input DNA sequence is tokenized and the output shape is verified.
    """
    mocker.patch.object(torch.cuda, "is_available", return_value=True)
    model = Caduceus(CaduceusConfig(model_name="caduceus-ph-4L-seqlen-1k-d118"))
    dataset = model.process_data([input_sequence])

    assert len(max(dataset["input_ids"], key=len)) <= model.config["input_size"]
    assert np.all(
        np.equal(np.array(expected_sequence), np.array(dataset["input_ids"][0]))
    )


@pytest.mark.skipif(caduceus_unavailable, reason="No Caduceus module present")
@pytest.mark.parametrize(
    "input_sequences, expected_sequences",
    [
        (
            ["A", "CC", "TTTT", "ACGTN", "ACG"],
            [
                [4, 4, 4, 4, 7, 1],
                [4, 4, 4, 8, 8, 1],
                [4, 10, 10, 10, 10, 1],
                [7, 8, 9, 10, 11, 1],
                [4, 4, 7, 8, 9, 1],
            ],
        )
    ],
)
def test_caduceus_process_data_variable_length_sequences(
    input_sequences, expected_sequences, mocker
):
    """
    Test the process_data method of the Caduceus model.
    The input DNA sequence is tokenized and the output shape is verified.
    """
    mocker.patch.object(torch.cuda, "is_available", return_value=True)
    model = Caduceus(CaduceusConfig(model_name="caduceus-ph-4L-seqlen-1k-d118"))
    dataset = model.process_data(input_sequences)

    assert len(max(dataset["input_ids"], key=len)) <= model.config["input_size"]
    assert np.all(
        np.equal(np.array(expected_sequences), np.array(dataset["input_ids"]))
    )



try:
    from helical.models.caduceus.fine_tuning_model import CaduceusFineTuningModel
    from helical.models.caduceus.model import CaduceusConfig
    caduceus_unavailable = False  # only run tests if able to import the package
except:
    caduceus_unavailable = True

from helical.models.fine_tune.fine_tuning_heads import RegressionHead
import os
import pytest

@pytest.mark.skipif(caduceus_unavailable, reason="No Caduceus module present")
def test_save_and_load_model():

    try:
        model = CaduceusFineTuningModel(CaduceusConfig(), "regression", 10)
        model.save_model("./caduceus_fine_tuned_model.pt")
        model.load_model("./caduceus_fine_tuned_model.pt")
        assert not model.model.training, "Model should be in eval mode"
        assert not model.fine_tuning_head.training, "Fine-tuning head should be in eval mode"
        assert model.fine_tuning_head.output_size == 10, "Output size should be 10"
        assert type(model.fine_tuning_head) == RegressionHead, "Fine-tuning head should be a RegressionHead"
        assert model.model is not None
        assert model.model.state_dict() is not None
    finally:
        if os.path.exists("./caduceus_fine_tuned_model.pt"):
            os.remove("./caduceus_fine_tuned_model.pt")



from helical.models.base_models import HelicalRNAModel
import logging
from pathlib import Path
import numpy as np
from anndata import AnnData
from helical.utils.downloader import Downloader
from transformers import BertForMaskedLM
from helical.models.geneformer.geneformer_utils import get_embs, quant_layers
from helical.models.geneformer.geneformer_tokenizer import TranscriptomeTokenizer
from helical.models.geneformer.geneformer_config import GeneformerConfig
from helical.utils.mapping import map_gene_symbols_to_ensembl_ids
from datasets import Dataset
from typing import Optional

LOGGER = logging.getLogger(__name__)


class Geneformer(HelicalRNAModel):
    """Geneformer Model.
    The Geneformer Model is a transformer-based model that can be used to extract gene embeddings from single-cell RNA-seq data. Both versions are made available through this interface.
    Both versions of Geneformer (v1 and v2) have different sub-models with varying numbers of layers, context size and pretraining set. The available models are the following:

    Version 1.0:
    - gf-12L-40M-i2048
    - gf-6L-10M-i2048

    Version 2.0:
    - gf-12L-38M-i4096
    - gf-12L-38M-i4096-CLcancer
    - gf-20L-151M-i4096

    Version 3.0:
    - gf-12L-104M-i4096
    - gf-12L-104M-i4096-CLcancer
    - gf-18L-316M-i4096

    For a detailed explanation of the differences between these models and versions, please refer to the Geneformer model card: https://helical.readthedocs.io/en/latest/model_cards/geneformer/

    Example
    -------
    ```python
    from helical.models.geneformer import Geneformer, GeneformerConfig
    import anndata as ad

    # Example configuration
    model_config = GeneformerConfig(model_name="gf-12L-38M-i4096", batch_size=10)
    geneformer_v2 = Geneformer(model_config)

    # Example usage for base pretrained model
    ann_data = ad.read_h5ad("anndata_file.h5ad")
    dataset = geneformer_v2.process_data(ann_data)
    embeddings = geneformer_v2.get_embeddings(dataset)
    print("Base model embeddings shape:", embeddings.shape)

    # Example usage for cancer-tuned model
    model_config_cancer = GeneformerConfig(model_name="gf-12L-38M-i4096-CLcancer", batch_size=10)
    geneformer_v2_cancer = Geneformer(model_config)

    cancer_ann_data = ad.read_h5ad("anndata_file.h5ad")
    cancer_dataset = geneformer_v2_cancer.process_data(cancer_ann_data)
    cancer_embeddings = geneformer_v2_cancer.get_embeddings(cancer_dataset)
    print("Cancer-tuned model embeddings shape:", cancer_embeddings.shape)
    ```

    Parameters
    ----------
    configurer : GeneformerConfig, optional, default = default_configurer
        The model configration

    Notes
    -----
    The first version of the model is published in this <a href="https://www.nature.com/articles/s41586-023-06139-9.epdf?sharing_token=u_5LUGVkd3A8zR-f73lU59RgN0jAjWel9jnR3ZoTv0N2UB4yyXENUK50s6uqjXH69sDxh4Z3J4plYCKlVME-W2WSuRiS96vx6t5ex2-krVDS46JkoVvAvJyWtYXIyj74pDWn_DutZq1oAlDaxfvBpUfSKDdBPJ8SKlTId8uT47M%3D">Nature Paper</a>.
    The second version of the model is available at <a href="https://pubmed.ncbi.nlm.nih.gov/39229018/">NIH</a>.
    We use the implementation from the <a href="https://huggingface.co/ctheodoris/Geneformer/tree/main">Geneformer</a> repository.

    """

    default_configurer = GeneformerConfig()

    def __init__(self, configurer: GeneformerConfig = default_configurer) -> None:
        super().__init__()
        self.configurer = configurer
        self.config = configurer.config
        self.files_config = configurer.files_config
        self.device = self.config["device"]

        downloader = Downloader()
        for file in self.configurer.list_of_files_to_download:
            downloader.download_via_name(file)

        self.model = BertForMaskedLM.from_pretrained(
            self.files_config["model_files_dir"],
            output_hidden_states=True,
            output_attentions=False,
        )
        self.model.eval()
        self.model = self.model.to(self.device)

        num_available_layers = quant_layers(self.model)
        available_layers = list(np.arange(num_available_layers))
        available_layers.append(-1)
        if self.config["emb_layer"] not in available_layers:
            message = f"Layer {self.config['emb_layer']} is not available. Available layers are {available_layers}."
            LOGGER.error(message)
            raise ValueError(message)

        if (
            self.config["emb_layer"] < (num_available_layers - 1)
            and self.config["emb_layer"] != -1
        ):
            message = f"Layer {self.config['emb_layer']} is not the last layer. This can lead to different embeddings and fine-tuning results."
            LOGGER.warning(message)

        self.layer_to_quant = self.config["emb_layer"]

        self.emb_mode = self.config["emb_mode"]
        self.forward_batch_size = self.config["batch_size"]

        self.tk = TranscriptomeTokenizer(
            custom_attr_name_dict=self.config["custom_attr_name_dict"],
            nproc=self.config["nproc"],
            model_input_size=self.config["input_size"],
            special_token=self.config["special_token"],
            gene_median_file=self.files_config["gene_median_path"],
            token_dictionary_file=self.files_config["token_path"],
            gene_mapping_file=self.files_config["ensembl_dict_path"],
        )

        self.pad_token_id = self.tk.gene_token_dict["<pad>"]
        self.cls_present = True if "<cls>" in self.tk.gene_token_dict else False
        self.eos_present = True if "<eos>" in self.tk.gene_token_dict else False

        LOGGER.info(f"Model finished initializing.")
        mode = "training" if self.model.training else "eval"
        LOGGER.info(
            f"'{self.config['model_name']}' model is in '{mode}' mode, on device '{self.device}' with embedding mode '{self.emb_mode}'."
        )

    def process_data(
        self,
        adata: AnnData,
        gene_names: str = "index",
        output_path: Optional[str] = None,
        use_raw_counts: bool = True,
    ) -> Dataset:
        """
        Processes the data for the Geneformer model.

        Parameters
        ----------
        adata : AnnData
            The AnnData object containing the data to be processed. Geneformer uses Ensembl IDs to identify genes
            and currently supports only human genes. If the AnnData object already has an 'ensembl_id' column,
            the mapping step can be skipped.
        gene_names : str, optional, default="index"
            The column in `adata.var` that contains the gene names. If set to a value other than "ensembl_id",
            the gene symbols in that column will be mapped to Ensembl IDs using the 'pyensembl' package,
            which retrieves mappings from the Ensembl FTP server and loads them into a local database.
            - If set to "index", the index of the AnnData object will be used and mapped to Ensembl IDs.
            - If set to "ensembl_id", no mapping will occur.
            Special case:
                If the index of `adata` already contains Ensembl IDs, setting this to "index" will result in
                invalid mappings. In such cases, create a new column containing Ensembl IDs and pass "ensembl_id"
                as the value of `gene_names`.
        output_path : str, optional, default=None
            If specified, saves the tokenized dataset to the given output path.
        use_raw_counts : bool, optional, default=True
            Determines whether raw counts should be used.

        Returns
        -------
        Dataset
            The tokenized dataset in the form of a Huggingface Dataset object.
        """
        LOGGER.info(f"Processing data for Geneformer.")
        self.ensure_rna_data_validity(adata, gene_names, use_raw_counts)

        # map gene symbols to ensemble ids if provided
        if gene_names != "ensembl_id":
            if (adata.var[gene_names].str.startswith("ENS").all()) or (
                adata.var[gene_names].str.startswith("None").any()
            ):
                message = (
                    "It seems an anndata with 'ensemble ids' and/or 'None' was passed. "
                    "Please set gene_names='ensembl_id' and remove 'None's to skip mapping."
                )
                LOGGER.info(message)
                raise ValueError(message)
            adata = map_gene_symbols_to_ensembl_ids(adata, gene_names)

            if adata.var["ensembl_id"].isnull().all():
                message = "All gene symbols could not be mapped to Ensembl IDs. Please check the input data."
                LOGGER.info(message)
                raise ValueError(message)

        tokenized_cells, cell_metadata = self.tk.tokenize_anndata(adata)

        # tokenized_cells, cell_metadata =  self.tk.tokenize_anndata(adata)
        tokenized_dataset = self.tk.create_dataset(
            tokenized_cells, cell_metadata, use_generator=False
        )

        if output_path:
            output_path = Path(output_path).with_suffix(".dataset")
            tokenized_dataset.save_to_disk(output_path)

        LOGGER.info(f"Successfully processed the data for Geneformer.")
        return tokenized_dataset

    def get_embeddings(
        self, dataset: Dataset, output_attentions: bool = False, output_genes: bool = False
    ) -> np.array:
        """Gets the gene embeddings from the Geneformer model

        Parameters
        ----------
        dataset : Dataset
            The tokenized dataset containing the processed data
        output_attentions : bool, optional, default=False
            Whether to output attentions from the model. This is useful for debugging or analysis purposes.
            The attention maps are averaged across heads and use the same layer as the embeddings.
        output_genes : bool, optional, default=False
            Whether to output the genes corresponding to the embeddings.

        Returns
        -------
        np.array
            The gene embeddings in the form of a numpy array
        list, optional
            The attention weights from the model, if `output_attentions` is set to True.
            The shape of the attention weights is (batch_size, num_heads, seq_length, seq_length).
            If `output_attentions` is False, this will not be returned.
        list, optional
            The list of genes corresponding to the embeddings, if `output_genes` is set to True.
            Each element in the list corresponds to the genes for each input in the dataset.
            If `output_genes` is False, this will not be returned.
        """
        LOGGER.info(f"Started getting embeddings:")
        embeddings = get_embs(
            self.model,
            dataset,
            self.emb_mode,
            self.layer_to_quant,
            self.pad_token_id,
            self.forward_batch_size,
            self.tk.gene_token_dict,
            self.tk.token_to_ensembl_dict,
            self.cls_present,
            self.eos_present,
            self.device,
            output_attentions=output_attentions,
            output_genes=output_genes,
        )

        LOGGER.info(f"Finished getting embeddings.")
        return embeddings



from typing import List, Union
from helical.models.base_models import HelicalDNAModel
from helical.utils.downloader import Downloader
from .caduceus_config import CaduceusConfig
from .pretrained_config import CaduceusPretrainedConfig
from .modeling_caduceus import CaduceusModel
from .caduceus_tokenizer import CaduceusTokenizer
from datasets import Dataset
import torch
from torch.utils.data import DataLoader
import numpy as np
import logging
from pandas import DataFrame

LOGGER = logging.getLogger(__name__)

# default configuration if not specified
configurer = CaduceusConfig()


class Caduceus(HelicalDNAModel):
    """Caduceus model.

    This class represents the Caduceus model, a DNA model using bi-directional and reverse complement DNA properties for better genomic analysis.

    Example
    ----------
    ```python
    from helical.models.caduceus import Caduceus, CaduceusConfig

    caduceus_config = CaduceusConfig(model_name="caduceus-ph-4L-seqlen-1k-d118", batch_size=5)
    caduceus = Caduceus(configurer = caduceus_config)

    sequence = ['ACTG' * int(1024/4)]
    processed_data = caduceus.process_data(sequence)

    embeddings = caduceus.get_embeddings(processed_data)
    print(embeddings.shape)
    ```

    Parameters
    ----------
    configurer : CaduceusConfig, optional, default=configurer
        The model configuration.

    Notes
    ----------
    This model has dependencies which only allow it to be run on CUDA devices.
    The link to the paper can be found [here](https://arxiv.org/abs/2403.03234).
    We make use of the implementation from the [Caduceus](https://github.com/kuleshov-group/caduceus) repository.
    """

    def __init__(self, configurer: CaduceusConfig = configurer):
        super().__init__()

        if torch.cuda.is_available() == False:
            message = "Caduceus requires a CUDA device to run and CUDA is not available"
            LOGGER.error(message)
            raise RuntimeError(message)

        self.configurer = configurer
        self.config = configurer.config
        self.files_config = configurer.files_config
        self.device = self.config["device"]

        downloader = Downloader()
        for file in self.configurer.list_of_files_to_download:
            downloader.download_via_name(file)

        self.pretrained_config = CaduceusPretrainedConfig.from_pretrained(
            self.files_config["model_files_dir"]
        )
        self.model = CaduceusModel.from_pretrained(
            self.files_config["model_files_dir"], config=self.pretrained_config
        )

        self.model.eval()

        self.tokenizer = CaduceusTokenizer(model_max_length=self.config["input_size"])

        LOGGER.info("Caduceus model initialized")

    def _collate_fn(self, batch):
        input_ids = torch.tensor([item["input_ids"] for item in batch])
        batch_dict = {
            "input_ids": input_ids,
        }

        if "labels" in batch[0]:
            batch_dict["labels"] = torch.tensor([item["labels"] for item in batch])

        return batch_dict

    def _pool_hidden_states(self, hidden_states, sequence_length_dim=1):
        """Pools hidden states along sequence length dimension."""
        if (
            self.config["pooling_strategy"] == "mean"
        ):  # Mean pooling along sequence length dimension
            return hidden_states.mean(dim=sequence_length_dim)
        if (
            self.config["pooling_strategy"] == "max"
        ):  # Max pooling along sequence length dimension
            return hidden_states.max(dim=sequence_length_dim).values
        if (
            self.config["pooling_strategy"] == "last"
        ):  # Use embedding of last token in the sequence
            return hidden_states.moveaxis(hidden_states, sequence_length_dim, 0)[
                -1, ...
            ]
        if (
            self.config["pooling_strategy"] == "first"
        ):  # Use embedding of first token in the sequence
            return hidden_states.moveaxis(hidden_states, sequence_length_dim, 0)[0, ...]

    def process_data(
        self,
        sequences: Union[List[str], DataFrame],
        return_tensors: str = "pt",
        padding: str = "max_length",
        truncation: bool = True,
    ) -> Dataset:
        """Process the input DNA sequences.

        Parameters
        ----------
        sequences : list[str] or DataFrame
            The input DNA sequences to be processed. If a DataFrame is provided, it should have a column named 'Sequence'.
        return_tensors : str, optional, default="pt"
            The return type of the processed data.
        padding : str, optional, default="max_length"
            The padding strategy to be used.
        truncation : bool, optional, default=True
            Whether to truncate the sequences or not.

        Returns
        ----------
        Dataset
            Containing processed DNA sequences.

        """
        LOGGER.info("Processing data for Caduceus.")
        sequences = self.get_valid_dna_sequence(sequences)

        max_length = min(len(max(sequences, key=len)), self.config["input_size"]) + 1

        # tokenized_sequences = []
        # for seq in sequences:
        tokenized_sequences = self.tokenizer(
            sequences,
            return_tensors=return_tensors,
            padding=padding,
            truncation=truncation,
            max_length=max_length,
        )
        # tokenized_sequences.append(tokenized_seq)

        dataset = Dataset.from_dict(tokenized_sequences)
        LOGGER.info("Successfully processed the data for Caduceus.")
        return dataset

    def get_embeddings(self, dataset: Dataset) -> np.ndarray:
        """Get the embeddings for the tokenized sequence.

        Parameters
        ----------
        dataset : Dataset
            The output dataset from `process_data`.

        Returns
        ----------
        np.ndarray
            The embeddings for the tokenized sequence in the form of a numpy array.
            NOTE: This method returns the embeddings using the pooling strategy specified in the config.
        """
        LOGGER.info("Started getting embeddings:")
        dataloader = DataLoader(
            dataset,
            collate_fn=self._collate_fn,
            batch_size=self.config["batch_size"],
            shuffle=False,
            num_workers=self.config["nproc"],
        )

        embeddings = []
        self.model.to(self.device)
        self.model.eval()
        for batch in dataloader:
            with torch.no_grad():
                outputs = self.model(input_ids=batch["input_ids"].to(self.device))
                embeddings.append(
                    self._pool_hidden_states(outputs.last_hidden_state).cpu().numpy()
                )

                del batch
                del outputs

        LOGGER.info(f"Finished getting embeddings.")
        return np.vstack(embeddings)



import numpy as np
from typing import List, Tuple, Union
from Bio.Seq import Seq
from tqdm import tqdm

import torch
from vortex.model.model import StripedHyena


def prepare_batch(
        seqs: List[str],
        tokenizer: object,
        prepend_bos: bool = False,
        device: str = 'cuda:0'
) -> Tuple[torch.Tensor, List[int]]:
    """
    Takes in a list of sequences, tokenizes them, and puts them in a tensor batch.
    If the sequences have differing lengths, then pad up to the maximum sequence length.
    """
    seq_lengths = [ len(seq) for seq in seqs ]
    max_seq_length = max(seq_lengths)

    input_ids = []
    for seq in seqs:
        padding = [tokenizer.pad_id] * (max_seq_length - len(seq))
        input_ids.append(
            torch.tensor(
                ([tokenizer.eod_id] * int(prepend_bos)) + tokenizer.tokenize(seq) + padding,
                dtype=torch.long,
            ).to(device).unsqueeze(0)
        )
    input_ids = torch.cat(input_ids, dim=0)

    return input_ids, seq_lengths


def logits_to_logprobs(
        logits: torch.Tensor,
        input_ids: torch.Tensor,
) -> torch.Tensor:
    """
    Takes in a tensor of logits of dimension (batch, length, vocab).
    Computes the log-likelihoods using a softmax along the vocab dimension.
    Uses the `input_ids` to index into the log-likelihoods and returns the likelihood
    of the provided sequence at each position with dimension (batch, length).
    """
    softmax_logprobs = torch.log_softmax(logits, dim=-1)
    softmax_logprobs = softmax_logprobs[:, :-1]
    input_ids = input_ids[:, 1:]
    assert softmax_logprobs.shape[1] == input_ids.shape[1]

    logprobs = torch.gather(
        softmax_logprobs,       # Gather likelihoods...
        2,                      # along the vocab dimension...
        input_ids.unsqueeze(-1) # using the token ids to index.
    ).squeeze(-1)

    return logprobs


def _score_sequences(
        seqs: List[str],
        model: StripedHyena,
        tokenizer: object,
        prepend_bos: bool = False,
        reduce_method: str = 'mean',
        device: str = 'cuda:0',
) -> List[float]:
    """Helper function to score a list of sequences based on their logprobs."""
    input_ids, seq_lengths = prepare_batch(seqs, tokenizer, device=device, prepend_bos=prepend_bos)
    assert len(seq_lengths) == input_ids.shape[0]

    with torch.inference_mode():
        logits, _ = model(input_ids) # (batch, length, vocab)

    logprobs = logits_to_logprobs(logits, input_ids)
    logprobs = logprobs.float().cpu().numpy()

    if reduce_method == 'sum': # PLL
        reduce_func = np.sum
    elif reduce_method == 'mean': # mean PLL
        reduce_func = np.mean
    else:
        raise ValueError(f'Invalid reduce_method {reduce_method}')

    return [
        reduce_func(logprobs[idx][:seq_lengths[idx]])
        for idx in range(len(seq_lengths))
    ]


def score_sequences(
        seqs: List[str],
        model: StripedHyena,
        tokenizer: object,
        batch_size: int = None,
        prepend_bos: bool = False,
        reduce_method: str = 'mean',
        device: str = 'cuda:0',
) -> List[float]:
    """
    Computes the model log-likelihood scores for sequences in `seqs`.
    Uses `reduce_method` to take the mean or sum across the likelihoods at each 
    position (default: `'mean'`).

    Returns a list of scalar scores corresponding to the reduced log-likelihoods for
    each sequence.
    """
    if batch_size is None:
        batch_size = len(seqs)

    scores = []
    for i in tqdm(range(0, len(seqs), batch_size)):
        batch_seqs = seqs[i:i + batch_size]
        batch_scores = _score_sequences(
            batch_seqs,
            model,
            tokenizer,
            prepend_bos=prepend_bos,
            reduce_method=reduce_method,
            device=device,
        )
        scores.extend(batch_scores)
    return scores


def score_sequences_rc(
        seqs: List[str],
        model: StripedHyena,
        tokenizer: object,
        batch_size: int,
        prepend_bos: bool = False,
        reduce_method: str = 'mean',
        device: str = 'cuda:0',
) -> List[float]:
    """
    Computes the model log-likelihood scores for sequences in `seqs` and for their
    reverse complements.
    Takes the mean score for the forward and reverse-complemented sequence.
    Uses `reduce_method` to take the mean or sum across the likelihoods at each 
    position (default: `'mean'`).

    Returns a list of scalar scores corresponding to the reduced log-likelihoods for
    each sequence.
    """
    scores = []
    for i in tqdm(range(0, len(seqs), batch_size)):
        batch_seqs = seqs[i:i + batch_size]
        batch_seqs_rc = [ str(Seq(seq).reverse_complement()) for seq in batch_seqs ]
        
        batch_scores = _score_sequences(
            batch_seqs,
            model,
            tokenizer,
            prepend_bos=prepend_bos,
            reduce_method=reduce_method,
            device=device,
        )
        batch_scores_rc = _score_sequences(
            batch_seqs_rc,
            model,
            tokenizer,
            prepend_bos=prepend_bos,
            reduce_method=reduce_method,
            device=device,
        )
        batch_scores = (np.array(batch_scores) + np.array(batch_scores_rc)) * 0.5

        scores.extend(list(batch_scores))
    return scores


def positional_entropies(
        seqs: List[str],
        model: StripedHyena,
        tokenizer: object,
        prepend_bos: bool = False,
        device: str = 'cuda:0',
) -> List[np.array]:
    """
    Computes the positional entropies for sequences in `seqs`.

    Returns a list of arrays, where each array is the same length as the
    corresponding sequence length. Each array contains the per-position entropy
    across the vocab dimension.
    """
    input_ids, seq_lengths = prepare_batch(seqs, tokenizer, device=device, prepend_bos=prepend_bos)
    assert len(seq_lengths) == input_ids.shape[0]

    with torch.inference_mode():
        logits, _ = model(input_ids) # (batch, length, vocab)
    
    softmax_logprobs = torch.log_softmax(logits, dim=-1)
    if prepend_bos:
        softmax_logprobs = softmax_logprobs[:, 1:, :] # Remove BOS entropy.

    entropies = -torch.sum(torch.exp(softmax_logprobs) * softmax_logprobs, dim=-1)
    entropies = entropies.float().cpu().numpy()

    sequence_entropies = [
        entropies[idx][:seq_lengths[idx]] for idx in range(len(seq_lengths))
    ]
    assert all(
        len(seq) == len(entropy) for seq, entropy in zip(seqs, sequence_entropies)
    )

    return sequence_entropies


def score_perplexity_along_sequence(
        model: StripedHyena,
        seq: str,
        reverse_complement: bool = True,
        entropy: bool = False
    ) -> np.array:
    '''
    Get forward and reverse RC of dna sequence, pass both through model, and return average entropy or perplexity.
    '''
    seq_rc = str(Seq(seq).reverse_complement())
    
    entropy_forward = positional_entropies([seq], model.model, model.tokenizer)[0]

    if reverse_complement:
        entropy_reverse = positional_entropies([seq_rc], model.model, model.tokenizer)[0]
        entropy_reverse = entropy_reverse[::-1]
        
        average_entropy = (entropy_forward + entropy_reverse) / 2
    else:
        average_entropy = entropy_forward

    if entropy:
        return average_entropy
    else:
        return np.exp(average_entropy)


import pytest
import numpy as np
from unittest import mock
from anndata import AnnData
from anndata._core.views import ArrayView
from pathlib import Path
import scipy.sparse as sp
from helical.models.geneformer.geneformer_tokenizer import (
    TranscriptomeTokenizer,
    sum_ensembl_ids,
)
from helical.constants.paths import CACHE_DIR_HELICAL


class TestGeneformerTokenizer:
    model_dir_v1 = Path(CACHE_DIR_HELICAL, "geneformer", "v1")
    model_dir_v2 = Path(CACHE_DIR_HELICAL, "geneformer", "v2")

    files_config_v1 = {
        "gene_median_path": model_dir_v1 / "gene_median_dictionary.pkl",
        "token_path": model_dir_v1 / "token_dictionary.pkl",
        "gene_mapping_path": model_dir_v1 / "ensembl_mapping_dict.pkl",
    }

    files_config_v2 = {
        "gene_median_path": model_dir_v2 / "gene_median_dictionary.pkl",
        "token_path": model_dir_v2 / "token_dictionary.pkl",
        "gene_mapping_path": model_dir_v2 / "ensembl_mapping_dict.pkl",
    }

    @pytest.fixture
    def tokenizer_v1(self):
        return TranscriptomeTokenizer(
            gene_median_file=self.files_config_v1["gene_median_path"],
            token_dictionary_file=self.files_config_v1["token_path"],
            gene_mapping_file=self.files_config_v1["gene_mapping_path"],
            model_input_size=2048,
            special_token=False,
        )

    @pytest.fixture
    def tokenizer_v2(self):
        return TranscriptomeTokenizer(
            gene_median_file=self.files_config_v2["gene_median_path"],
            token_dictionary_file=self.files_config_v2["token_path"],
            gene_mapping_file=self.files_config_v2["gene_mapping_path"],
            model_input_size=4096,
            special_token=True,
        )

    def test_tokenizer_initialization(self, tokenizer_v1, tokenizer_v2):
        assert tokenizer_v1.model_input_size == 2048
        assert not tokenizer_v1.special_token
        assert tokenizer_v2.model_input_size == 4096
        assert tokenizer_v2.special_token

    @pytest.mark.parametrize(
        "collapse_gene_ids, expected_shape, expected_exception, data_format",
        [
            (True, (2, 2), None, "h5ad"),
            (False, (2, 3), ValueError, "h5ad"),
            (True, (2, 3), ValueError, "bad_format"),
        ],
    )
    def test_sum_ensembl_ids_h5ad(
        self, collapse_gene_ids, expected_shape, expected_exception, data_format
    ):
        # Create a test AnnData object
        adata = AnnData(X=sp.csr_matrix([[1, 2, 3], [4, 5, 6]]))
        adata.var["ensembl_id"] = ["ENSG1", "ENSG2", "ENSG2"]
        adata.obs["n_counts"] = [6, 15]

        # Mock gene_mapping_dict and gene_token_dict
        gene_mapping_dict = {"ENSG1": "ENSG1", "ENSG2": "ENSG2"}
        gene_token_dict = {"ENSG1": 1, "ENSG2": 2}

        if expected_exception:
            with pytest.raises(expected_exception):
                sum_ensembl_ids(
                    adata,
                    collapse_gene_ids,
                    gene_mapping_dict,
                    gene_token_dict,
                    file_format=data_format,
                )
        else:
            result = sum_ensembl_ids(
                adata,
                collapse_gene_ids,
                gene_mapping_dict,
                gene_token_dict,
                file_format=data_format,
            )
            assert isinstance(result, AnnData)
            assert result.shape == expected_shape

    def test_tokenize_anndata_sparse_matrix(self, tokenizer_v1):
        # Create a test AnnData object
        adata = AnnData(X=sp.csr_matrix([[1, 2, 3], [4, 5, 6]]))
        adata.var["ensembl_id"] = ["ENSG1", "ENSG2", "ENSG3"]
        adata.obs["n_counts"] = [6, 15]
        adata.obs["cell_type"] = ["A", "B"]

        tokenizer_v1.custom_attr_name_dict = {"cell_type": "cell_type"}
        tokenized_cells, cell_metadata = tokenizer_v1.tokenize_anndata(adata)

        assert len(tokenized_cells) == 2
        assert isinstance(tokenized_cells[0], np.ndarray)
        assert cell_metadata == {"cell_type": ["A", "B"]}

    def test_tokenize_anndata_sparse_matrix_without_n_counts(self, tokenizer_v1):
        # Test the case where n_counts is not present in the AnnData object and
        # it's estimated from the sum of the counts in the sparse matrix

        # Create a test AnnData object
        adata = AnnData(X=sp.csr_matrix([[1, 2, 3], [4, 5, 6]]))
        adata.var["ensembl_id"] = ["ENSG1", "ENSG2", "ENSG3"]
        adata.obs["n_counts"] = [6, 15]
        adata.obs["cell_type"] = ["A", "B"]

        tokenizer_v1.custom_attr_name_dict = {"cell_type": "cell_type"}
        tokenized_cells, cell_metadata = tokenizer_v1.tokenize_anndata(adata)

        assert len(tokenized_cells) == 2
        assert isinstance(tokenized_cells[0], np.ndarray)
        assert cell_metadata == {"cell_type": ["A", "B"]}

    def test_tokenize_anndata_array_view(self, tokenizer_v1):
        # Test the case where the AnnData X is a dense matrix of type
        # ArrayView object
        adata = AnnData(X=ArrayView(np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)))
        adata.var["ensembl_id"] = ["ENSG1", "ENSG2", "ENSG3"]
        adata.obs["cell_type"] = ["A", "B"]

        tokenizer_v1.custom_attr_name_dict = {"cell_type": "cell_type"}
        tokenized_cells, cell_metadata = tokenizer_v1.tokenize_anndata(adata)

        assert len(tokenized_cells) == 2
        assert isinstance(tokenized_cells[0], np.ndarray)
        assert cell_metadata == {"cell_type": ["A", "B"]}

    @pytest.mark.parametrize(
        "tokenizer, expected_max_length",
        [("tokenizer_v1", 2048), ("tokenizer_v2", 4096)],
    )
    def test_create_dataset(self, request, tokenizer, expected_max_length):
        tokenizer = request.getfixturevalue(tokenizer)
        tokenized_cells = [np.array([1, 2, 3]), np.array([4, 5, 6])]
        cell_metadata = None

        dataset = tokenizer.create_dataset(tokenized_cells, cell_metadata)
        assert len(dataset) == 2
        assert "input_ids" in dataset.features
        assert "length" in dataset.features
        assert all(len(ids) <= expected_max_length for ids in dataset["input_ids"])

        if tokenizer.special_token:
            assert all(
                ids[0] == tokenizer.gene_token_dict["<cls>"]
                for ids in dataset["input_ids"]
            )
            assert all(
                ids[-1] == tokenizer.gene_token_dict["<eos>"]
                for ids in dataset["input_ids"]
            )

    def test_tokenize_data(self, tokenizer_v1, tmp_path):
        # Create a test AnnData object
        adata = AnnData(X=sp.csr_matrix([[1, 2, 3], [4, 5, 6]]))
        adata.var["ensembl_id"] = ["ENSG1", "ENSG2", "ENSG3"]
        adata.obs["n_counts"] = [6, 15]
        adata.obs["cell_type"] = ["A", "B"]
        test_file = tmp_path / "test.h5ad"
        adata.write_h5ad(test_file)

        output_dir = tmp_path / "output"
        output_dir.mkdir()

        tokenizer_v1.tokenize_data(
            tmp_path, output_dir, "test_output", file_format="h5ad"
        )

        assert (output_dir / "test_output.dataset").exists()

        with pytest.raises(ValueError):
            tokenizer_v1.tokenize_data(
                tmp_path, output_dir, "test_output", file_format="bad_format"
            )



from helical.models.helix_mrna import HelixmRNA, HelixmRNAConfig
import hydra
from omegaconf import DictConfig
from pandas import DataFrame


@hydra.main(
    version_base=None,
    config_path="../run_models/configs",
    config_name="helix_mrna_config",
)
def run(cfg: DictConfig):
    input_sequences = DataFrame(
        {"Sequence": ["EACU" * 20, "EAUG" * 20, "EAUG" * 20, "EACU" * 20, "EAUU" * 20]}
    )

    helix_mrna_config = HelixmRNAConfig(**cfg)
    helix_mrna = HelixmRNA(helix_mrna_config)

    processed_input_data = helix_mrna.process_data(input_sequences)

    embeddings = helix_mrna.get_embeddings(processed_input_data)
    print(embeddings.shape)


if __name__ == "__main__":
    run()



###############################################################################
# Auto-generated by `jupyter-book config`
# If you wish to continue using _config.yml, make edits to that file and
# re-generate this one.
###############################################################################
author = "Helical Team"
autosummary_generate = True
bibtex_bibfiles = ["references.bib"]
comments_config = {"hypothesis": False, "utterances": False}
copyright = "2024"
exclude_patterns = ["**.ipynb_checkpoints", ".DS_Store", "Thumbs.db", "_build"]
extensions = [
    "sphinx_togglebutton",
    "sphinx_copybutton",
    "myst_nb",
    "jupyter_book",
    "sphinx_thebe",
    "sphinx_comments",
    "sphinx_external_toc",
    "sphinx.ext.intersphinx",
    "sphinx_design",
    "sphinx_book_theme",
    "sphinx_inline_tabs",
    "sphinx_proof",
    "sphinx_examples",
    "hoverxref.extension",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinxcontrib.bibtex",
    "sphinx_jupyterbook_latex",
    "sphinx_multitoc_numbering",
]
external_toc_exclude_missing = False
external_toc_path = "_toc.yml"
hoverxref_intersphinx = ["sphinxproof"]
html_baseurl = ""
html_css_files = ["my_theme copy.css", "my_theme.css"]
html_favicon = "docs/official_logo_fav.png"
html_logo = "docs/official_logo_small.png"
html_sourcelink_suffix = ""
html_static_path = ["_static"]
html_theme = "sphinx_book_theme"
html_theme_options = {
    "search_bar_text": "Search this book...",
    "launch_buttons": {
        "notebook_interface": "classic",
        "binderhub_url": "",
        "jupyterhub_url": "",
        "thebe": False,
        "colab_url": "",
    },
    "path_to_docs": "docs",
    "repository_url": "https://github.com/helicalAI/helical",
    "repository_branch": "main",
    "extra_footer": "",
    "home_page_in_toc": True,
    "announcement": "",
    "analytics": {"google_analytics_id": ""},
    "use_repository_button": True,
    "use_edit_page_button": False,
    "use_issues_button": True,
}
html_title = "Helical Documentation"
intersphinx_mapping = {
    "ebp": ["https://executablebooks.org/en/latest/", None],
    "myst-parser": ["https://myst-parser.readthedocs.io/en/latest/", None],
    "myst-nb": ["https://myst-nb.readthedocs.io/en/latest/", None],
    "sphinx": ["https://www.sphinx-doc.org/en/master", None],
    "nbformat": ["https://nbformat.readthedocs.io/en/latest", None],
    "sd": ["https://sphinx-design.readthedocs.io/en/latest", None],
    "sphinxproof": ["https://sphinx-proof.readthedocs.io/en/latest/", None],
}
latex_engine = "pdflatex"
mathjax3_config = {
    "tex": {
        "macros": {
            "N": "\\mathbb{N}",
            "floor": ["\\lfloor#1\\rfloor", 1],
            "bmat": ["\\left[\\begin{array}"],
            "emat": ["\\end{array}\\right]"],
        }
    }
}
myst_enable_extensions = [
    "colon_fence",
    "dollarmath",
    "linkify",
    "substitution",
    "tasklist",
]
myst_url_schemes = ["mailto", "http", "https"]
napoleon_include_init_with_doc = True
napoleon_numpy_docstring = True
nb_execution_allow_errors = False
nb_execution_cache_path = ""
nb_execution_excludepatterns = []
nb_execution_in_temp = False
nb_execution_mode = "off"
nb_execution_timeout = 30
nb_output_stderr = "show"
numfig = True
pygments_style = "sphinx"
suppress_warnings = ["myst.domains"]
use_jupyterbook_latex = True
use_multitoc_numbering = True



from helical.models.uce.fine_tuning_model import UCEFineTuningModel
from helical.models.uce.model import UCEConfig
from helical.models.fine_tune.fine_tuning_heads import ClassificationHead
import os

def test_save_and_load_model():

    try:
        model = UCEFineTuningModel(UCEConfig(), "classification", 10)
        model.save_model("./uce_fine_tuned_model.pt")
        model.load_model("./uce_fine_tuned_model.pt")
        assert not model.model.training, "Model should be in eval mode"
        assert not model.fine_tuning_head.training, "Fine-tuning head should be in eval mode"
        assert model.fine_tuning_head.output_size == 10, "Output size should be 10"
        assert type(model.fine_tuning_head) == ClassificationHead, "Fine-tuning head should be a ClassificationHead"
        assert model.model is not None
        assert model.model.state_dict() is not None
    finally:
        if os.path.exists("./uce_fine_tuned_model.pt"):
            os.remove("./uce_fine_tuned_model.pt")
